import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GraphicsConfiguration;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream.GetField;

import javax.imageio.ImageIO;

public class Export {
	private static void exportLanguageToUTF(final String inFile, final String outFile, final String inEnc) {
		BufferedReader brd = null;
		DataOutputStream dos = null;
		try {
			brd = new BufferedReader(new InputStreamReader( new FileInputStream(inFile), inEnc ) );
			dos = new DataOutputStream( new FileOutputStream( outFile ));

			String str;
			while( (str = brd.readLine()) != null ) {
				dos.writeUTF( str );
			}
		} catch(IOException e) {
		} finally {
			try{if (brd != null)brd.close();}catch(IOException ie){}
			try{if (dos != null)dos.close();}catch(IOException ie){}
		}
	}
	private static void exportCos() {
		final int ANGLE_360 = 65536;
		final int PRECISION = 16384;
		double angle = 0.0;
		double step = (2.0 * Math.PI) / (double)ANGLE_360;
		FileOutputStream fos = null;
		try {
			fos = new FileOutputStream( "cos" );
			for( int i = 0; i < ANGLE_360 / 4 + 1; i++ ) {
				short cos = (short)(Math.cos( angle ) * (double)PRECISION);
				angle += step;
				// big endian
				fos.write( (int) (cos >> 8) );
				fos.write( (int) cos );
			}
		} catch(IOException e) {
		} finally {
			if( fos != null ) try{ fos.close(); } catch( IOException e ) {}
		}
	}
	/**
	 * code1,c1,c2,...cn,code2,c1,c2,...cn
	 * @param fontName
	 * @param size
	 * @param charset
	 */
	private static void exportFont(final String fontName, int size, final char[] charset) {
		FileInputStream is = null;
		DataOutputStream os = null;
		try {
			is = new FileInputStream( fontName );
			final Font font = Font.createFont( Font.TRUETYPE_FONT, is ).deriveFont( Font.PLAIN, size );
			BufferedImage img = new BufferedImage( size * 2, 4 * 8, BufferedImage.TYPE_INT_ARGB_PRE );
			final Graphics g = img.getGraphics();
			g.setFont( font );
			final FontMetrics fm = g.getFontMetrics();
			final int char_width = fm.charWidth( ' ' ) + fm.getLeading() + 0;
			final int font_baseline = ((size + 1 + fm.getAscent() - fm.getDescent()) >> 1) + 0;
			final int size_mask = (1 << size) - 1;
			// precalculate
			final int[] counter = new int[16];
			final int[][] bit_font = new int[16][];
			for( int c = 0; c < charset.length; c++ ) {
				counter[charset[c] & 0xf]++;
			}
			for( int c = 0; c < counter.length; c++ ) {
				bit_font[c] = new int[counter[c] * (char_width + 1)];
				counter[c] = 0;
			}
			for( int c = 0; c < charset.length; c++ ) {
				g.setColor( Color.WHITE );
				g.fillRect( 0, 0, size * 2, 4 * 8 );
				g.setColor( Color.BLACK );
				g.drawChars( charset, c, 1, 0, font_baseline );
				//
				final int block = charset[c] & 0xf;
				bit_font[block][counter[block]++] = (int)charset[c];
				//
				System.out.println( charset[c] );
				for( int i = 0; i < char_width; i++ ) {
					int column = 0;
					for( int h = 0; h < 4 * 8; h++ ) {
						column >>= 1;
						column &= 0x7fffffff;
						int pixel = img.getRGB( i, h );
						//System.out.print( pixel );
						if( pixel == 0xff000000 ) {
							column |= 0x80000000;
						}
					}
					column &= size_mask;
					bit_font[block][counter[block]++] = column;
					//System.out.println();
					String str = Integer.toString( column, 2 );
					str = "00000000000000000000000000000000".substring( str.length() ) + str;
					System.out.println( str );
				}
				
			}
			g.dispose();
			img.flush();
			// save result
			os = new DataOutputStream( new FileOutputStream("font" + Integer.toString( size )) );
			os.writeByte( size );
			os.writeByte( char_width );
			os.writeByte( font_baseline );
			for( int block = 0; block < bit_font.length; block++ ) {
				final int[] char_block = bit_font[block];
				// block length
				os.writeInt( char_block.length );
				for( int d = 0; d < char_block.length; ) {
					os.writeChar( char_block[d++] );// char code
					for( int s = 0; s < char_width; s++ )
						os.writeInt( char_block[d++] );// bit mask
				}
			}
		} catch( Exception e ) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		} finally {
			if( is != null ) try { is.close(); } catch (IOException e) {}
			if( os != null ) try { os.close(); } catch (IOException e) {}
		}
	}
	private static void exportFontAsPicture(final String fontName, int size, int style, final char[] charset) {
		FileInputStream is = null;
		try {
			is = new FileInputStream( fontName );
			final Font font = Font.createFont( Font.TRUETYPE_FONT, is ).deriveFont( style, size );
			BufferedImage img = new BufferedImage( size * 2, 4 * 8, BufferedImage.TYPE_INT_ARGB_PRE );
			Graphics g = img.getGraphics();
			g.setFont( font );
			final FontMetrics fm = g.getFontMetrics();
			final int char_width = fm.charWidth( ' ' );// + fm.getLeading() + 0;
			final int font_baseline = ((size + 1 + fm.getAscent() - fm.getDescent()) >> 1) + 0;
			final int size_mask = (1 << size) - 1;
			// precalculate
			final int chars_in_line = 16;
			final int image_width = chars_in_line * char_width;
			final int image_height = size * ((charset.length / chars_in_line) + ((charset.length % chars_in_line) != 0 ? 1 : 0));
			g.dispose();
			img.flush();
			img = new BufferedImage( image_width, image_height, BufferedImage.TYPE_INT_ARGB_PRE );
			g = img.getGraphics();
			g.setFont( font );
			g.setColor( Color.WHITE );
			g.fillRect( 0, 0, image_width, image_height );
			// print chars
			g.setColor( Color.BLACK );
			for( int i = 0, y = font_baseline; i < charset.length; i += chars_in_line ) {
				int length = Math.min( chars_in_line, charset.length - i );
				g.drawChars( charset, i, length, 0, y );
				y += size;
				System.out.println( new String( charset, i, length ) );
			}
			//
			g.dispose();
			// save result
			ImageIO.write( img, "png", new File("font" + size + ".png") );
			img.flush();
		} catch( Exception e ) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		} finally {
			if( is != null ) try { is.close(); } catch (IOException e) {}
		}
	}
	private static void saveCharsetData(final char[] charset, int size, int width, int baseLine) {
		DataOutputStream os = null;
		try{
			os = new DataOutputStream( new FileOutputStream("charset") );
			os.writeByte( size );
			os.writeByte( width );
			os.writeByte( baseLine );
			for( int i = 0; i < charset.length; i++ ) {
				os.writeChar( charset[i] );
			}
		} catch( Exception e ) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		} finally {
			if( os != null ) try { os.close(); } catch (IOException e) {}
		}
	}
	private static void exportPictureAsFont(final String name, final char[] charset, int size, int char_width, int font_baseline) {
		FileInputStream is = null;
		DataOutputStream os = null;
		try {
			BufferedImage img = ImageIO.read( new File( name ) );
			final Graphics g = img.getGraphics();
			// precalculate
			final int[] counter = new int[16];
			final int[][] bit_font = new int[16][];
			for( int c = 0; c < charset.length; c++ ) {
				counter[charset[c] & 0xf]++;
			}
			for( int c = 0; c < counter.length; c++ ) {
				bit_font[c] = new int[counter[c] * (char_width + 1)];
				counter[c] = 0;
			}
			//
			final int char_in_line = img.getWidth() / char_width;
			for( int c = 0; c < charset.length; c++ ) {
				final int block = charset[c] & 0xf;
				bit_font[block][counter[block]++] = (int)charset[c];
				//
				System.out.println( charset[c] );
				final int x = (c % char_in_line) * char_width;
				final int y = (c / char_in_line) * size;
				for( int i = x; i < x + char_width; i++ ) {
					int column = 0;
					for( int h = 0; h < size; h++ ) {
						column >>= 1;
						column &= 0x7fffffff;
						int pixel = img.getRGB( i, h + y );
						//System.out.print( pixel );
						if( pixel == 0xff000000 ) {
							column |= 0x80000000;
						}
					}
					for( int h = size; h < 4 * 8; h++ ) {
						column >>= 1;
						column &= 0x7fffffff;
					}
					bit_font[block][counter[block]++] = column;
					//System.out.println();
					String str = Integer.toString( column, 2 );
					str = "00000000000000000000000000000000".substring( str.length() ) + str;
					System.out.println( str );
				}
				
			}
			g.dispose();
			img.flush();
			// save result
			os = new DataOutputStream( new FileOutputStream("font" + size + "x" + char_width) );
			os.writeByte( size );
			os.writeByte( char_width );
			os.writeByte( font_baseline );
			for( int block = 0; block < bit_font.length; block++ ) {
				final int[] char_block = bit_font[block];
				// block length
				os.writeInt( char_block.length );
				for( int d = 0; d < char_block.length; ) {
					os.writeChar( char_block[d++] );// char code
					for( int s = 0; s < char_width; s++ ) {
						//os.writeInt( char_block[d++] );// bit mask
						final int data = char_block[d++];
						if( size <= 8 )
							os.writeByte( data );
						else if( size <= 16 )
							os.writeShort( data );
						else
							os.writeInt( data );
					}
				}
			}
		} catch( Exception e ) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		} finally {
			if( is != null ) try { is.close(); } catch (IOException e) {}
			if( os != null ) try { os.close(); } catch (IOException e) {}
		}
	}
	private static final char[] CharSet =
		(" ',-.:/" +
		"0123456789" +
		">?" +
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ" +
		"abcdefghijklmnopqrstuvwxyz" +
		"АБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯ" +
		"абвгдеёжзийклмнопрстуфхцчшщъыьэюя").toCharArray();
	public static void main(String args[]) {
		//exportCos();
		//exportLanguageToUTF("../StuntCar/src/en.properties", "en.properties", "UTF-8");
		//exportLanguageToUTF("../StuntCar/src/ru.properties", "ru.properties", "UTF-8");
		//exportLanguageToUTF("en.txt", "en.properties", "UTF-8");
		//exportLanguageToUTF("ru.txt", "ru.properties", "UTF-8");
		//exportFont("../StuntCar/src/Lucida Console Regular.ttf", 7, CharSet );
		//exportFontAsPicture("../StuntCar/src/Lucida Console Regular.ttf", 7, Font.PLAIN, CharSet );
		//exportFontAsPicture("consola.ttf", 9, Font.PLAIN, CharSet );
		//saveCharsetData( CharSet, 7, 8, 6 );
		//exportPictureAsFont( "MyFont8.png", CharSet, 8, 5, 7 );
		exportPictureAsFont( "MyFont7.png", CharSet, 7, 5, 6 );
		//exportPictureAsFont( "MyFont7x4.png", CharSet, 7, 4, 6 );
	}
}