public class Camera extends Coord3D {
	/** precision for calculating trigonometric, 14 max */
	public static final int PRECISION_LOG2 = 14;
	public static final int PRECISION = 1 << PRECISION_LOG2;
	/** value that is equivalent to 360 degrees */
	public static final int ANGLE_360 = 65536;
	/** value that is equivalent to 180 degrees */
	public static final int ANGLE_180 = 180 * ANGLE_360 / 360;
	/** value that is equivalent to 90 degrees */
	public static final int ANGLE_90 = 90 * ANGLE_360 / 360;
	/** cosines for angles from 0 to 90 degrees */
	public static final short[] sCosTable = new short[ANGLE_90 + 1];
	static {
		//initCosTable
		double angle = 0.0;
		double step = (2.0 * Math.PI) / (double)ANGLE_360;
		for( int i = 0; i < Camera.sCosTable.length; i++ ) {
			sCosTable[i] = (short)(Math.cos( angle ) * (double)PRECISION);
			angle += step;
		}
		/*
		DataInputStream is = null;
		try {
			is = new DataInputStream( GameView.class.getResourceAsStream("cos") );
			for( int i = 0; i < mCosTable.length; i++ ) {
				mCosTable[i] = is.readShort();
			}
		} catch(Exception e) {
			System.out.println( e.getMessage() );
		} finally {
			if( is != null ) try { is.close(); } catch (IOException ex) { }
		}
		*/
	}
	//
	public int x_angle;
	public int sin_x;
	public int cos_x;
	//
	public int y_angle;
	public int sin_y;
	public int cos_y;
	//
	public int z_angle;
	public int sin_z;
	public int cos_z;
	/**
	 * Matrix to rotate coordinates, mr = 4<br>
	 * to rotate from world to local coordinates:<br>
	 * x = [0*mr+0] * x + [0*mr+1] * y + [0*mr+2] * z<br>
	 * y = [1*mr+0] * x + [1*mr+1] * y + [1*mr+2] * z<br>
	 * z = [2*mr+0] * x + [2*mr+1] * y + [2*mr+2] * z<br>
	 * to rotate from local to world coordinates:<br>
	 * x = [0*mr+0] * x + [1*mr+0] * y + [2*mr+0] * z<br>
	 * y = [0*mr+1] * x + [1*mr+1] * y + [2*mr+1] * z<br>
	 * z = [0*mr+2] * x + [1*mr+2] * y + [2*mr+2] * z<br>
	 */
	public int[] mMatrix = new int[3 * 3];

	public Camera() {
		this( 0, 0, 0 );
	}
	public Camera(int xcoord, int ycoord, int zcoord) {
		super( xcoord, ycoord, zcoord );
		setAngles( 0, 0, 0 );
	}
	public final void setAngles(int xangle, int yangle, int zangle) {
		setXangle( xangle );
		setYangle( yangle );
		setZangle( zangle );
	}
	public final void setXangle(int angle) {
		x_angle = (angle < -ANGLE_180) ? (ANGLE_360 + angle) : ((angle > ANGLE_180) ? (angle - ANGLE_360) : angle);
		angle = Math.abs( x_angle );
		if( angle <= ANGLE_90 ) {
			cos_x = (int)sCosTable[angle];
			sin_x = (int)sCosTable[ANGLE_90 - angle];
			if( x_angle < 0 ) sin_x = -sin_x;
			return;
		}
		cos_x = -(int)sCosTable[ANGLE_180 - angle];
		sin_x = (int)sCosTable[angle - ANGLE_90];
		if( x_angle < 0 ) sin_x = -sin_x;
	}
	public final void setYangle(int angle) {
		y_angle = (angle < -ANGLE_180) ? (ANGLE_360 + angle) : ((angle > ANGLE_180) ? (angle - ANGLE_360) : angle);
		angle = Math.abs( y_angle );
		if( angle <= ANGLE_90 ) {
			cos_y = (int)sCosTable[angle];
			sin_y = (int)sCosTable[ANGLE_90 - angle];
			if( y_angle < 0 ) sin_y = -sin_y;
			return;
		}
		cos_y = -(int)sCosTable[ANGLE_180 - angle];
		sin_y = (int)sCosTable[angle - ANGLE_90];
		if( y_angle < 0 ) sin_y = -sin_y;
	}
	public final void setZangle(int angle) {
 		z_angle = (angle < -ANGLE_180) ? (ANGLE_360 + angle) : ((angle > ANGLE_180) ? (angle - ANGLE_360) : angle);
		angle = Math.abs( z_angle );
		if( angle <= ANGLE_90 ) {
			cos_z = (int)sCosTable[angle];
			sin_z = (int)sCosTable[ANGLE_90 - angle];
			if( z_angle < 0 ) sin_z = -sin_z;
			return;
		}
		cos_z = -(int)sCosTable[ANGLE_180 - angle];
		sin_z = (int)sCosTable[angle - ANGLE_90];
		if( z_angle < 0 ) sin_z = -sin_z;
	}
	public final void calculateZXYmatrix() {
		// around z
		//x = x * cos z - y * sin z
		//y = y * cos z + x * sin z
		// around x
		//y = y * cos x - z * sin x
		//z = z * cos x + y * sin x
		// around y
		//x = x * cos y + z * sin y
		//z = z * cos y - x * sin y
		//
		final int cos_y_cos_z = (cos_y * cos_z) / PRECISION;
		final int sin_x_sin_y = (sin_x * sin_y) / PRECISION;
		final int cos_y_sin_z = (cos_y * sin_z) / PRECISION;
		final int[] m = mMatrix;
		// x = x * cos z * cos y + x * sin z * sin x * sin y + y * cos z * sin x * sin y - y * sin z * cos y + z * cos x * sin y
		m[0] = cos_y_cos_z + (sin_x_sin_y * sin_z) / PRECISION;
		m[1] = (sin_x_sin_y * cos_z) / PRECISION - cos_y_sin_z;
		m[2] = (cos_x * sin_y) / PRECISION;
		// y = x * sin z * cos x + y * cos z * cos x - z * sin x
		m[3] = (sin_z * cos_x) / PRECISION;
		m[4] = (cos_z * cos_x) / PRECISION;
		m[5] = - sin_x;
		// z = x * sin z * sin x * cos y - x * cos z * sin y + y * cos z * sin x * cos y + y * sin z * sin y + z * cos x * cos y
		m[6] = (cos_y_sin_z * sin_x - cos_z * sin_y) / PRECISION;
		m[7] = (cos_y_cos_z * sin_x + sin_z * sin_y) / PRECISION;
		m[8] = (cos_x * cos_y) / PRECISION;
	}
	/**
	 * @param direction - +1 clockwise, -1 counter clockwise
	 * @param center x and y - coordinates of a pivot point
	 */
	public void rotate(final int direction, final int center_x, final int center_y) {
		setZangle( z_angle + direction * ANGLE_90 );
		// around z
		//x = x * cos z - y * sin z
		//y = y * cos z + x * sin z
		// angle = -90 grad
		// sin z = -1
		// cos z = 0
		int x = mX - center_x;
		int y = mY - center_y;
		mX = center_x + direction * y;
		mY = center_y - direction * x;
	}
}