/**
 * mX, mY, mZ is car camera coordinates
 * Car CG coordinates is mCG_X_fp8, mCG_Y_fp8, mCG_Z_fp8 and mCG_X, mCG_Y, mCG_Z
 */
public class Car extends Camera {
	// earth gravity = fp8 * 981 px/s / k * (dt/1000)^2
	// game gravity = gravity / 2
	protected static final int GRAVITY_ACCELERATION_FP8 = -(int)(256 * 981 * GameView.REFRESH_PERIOD * GameView.REFRESH_PERIOD / 1000 / 1000 / 2);
	/** For one wheel. t to reach 100 km/h, 2778/t/4. */
	protected static final int ENGINE_ACCELERATION_FP8 = (int)(256 * 136 * GameView.REFRESH_PERIOD*GameView.REFRESH_PERIOD/1000/1000);
	protected static final int MAX_SPEED_FP8 = ((int)(256 * 160 * GameView.REFRESH_PERIOD) / 36);// < 160 kmh
	protected static final int DAMAGE_LIMIT = 300;
	/** Center Gravitation coordinates */
	// base point. CG_X = 0
	protected static final int CG_Y = 0;// cm
	protected static final int CG_Z = 50;// cm
	protected static final int WIDTH = 128;// cm
	protected static final int LENGTH = 256;// cm
	protected static final int HEIGHT = 120;// cm
	protected static final int WHEEL_RADIUS = 32;// cm
	protected static final int WHEEL_WIDTH = 16;// cm
	protected static final int CENTER_LINE_MARGIN = 30;// cm
	// wheels
	// springs, travel is 20 cm
	protected static final int SPRINT_TRAVEL = 20;// cm
	// coordinates relative to base point
	protected static final int WHEEL_LEFT_X = -(WIDTH / 2 - WHEEL_WIDTH / 2);
	protected static final int WHEEL_RIGHT_X = -WHEEL_LEFT_X;
	protected static final int WHEEL_FRONT_Y = LENGTH / 2 - WHEEL_RADIUS - CG_Y;
	protected static final int WHEEL_REAR_Y = -LENGTH / 2 + WHEEL_RADIUS - CG_Y;
	protected static final int WHEEL_FRONT_REAR_DISTANCE = WHEEL_FRONT_Y - WHEEL_REAR_Y;
	protected static final int WHEEL_TREAD = WHEEL_RIGHT_X - WHEEL_LEFT_X;
	protected static final int WHEEL_BOTTOM_Z = 0 - CG_Z;
	/** Camera Z position relative to base point */
	protected static final int LOOK_Z = (100 - CG_Z);
	/** Camera Y position relative to base point */
	protected static final int LOOK_Y = 0;
	//protected static final int LOOK_X = 0;
	// relative to base point
	protected static final int BONNET = LENGTH / 2 - WHEEL_RADIUS / 2 - CG_Y;// cm
	//
	public int mPieceIndex = 0;// current piece
	//public TrackPiece mPiece = null;
	public int mPieceCoordIndex = 0;// to identify segment
	public int mCoordIndex = 0;// to calculate distance
	public boolean mIsOnRoad = false;
	public boolean mIsOffRoad = false;
	public boolean mIsScratchSide = false;
	public boolean mIsTouchingRoad = false;
	public boolean mIsAccelerating = false;
	public boolean mIsBoostActivated = false;
	public int mLap = 0;
	public long mStartRaceTime = 0;
	public long mStartLapTime = 0;
	public int mBestLapTime = 0;
	public int mRaceTime = 0;
	public int mTravelDistance = 0;
	protected int mCG_X_fp8 = 0;
	protected int mCG_Y_fp8 = 0;
	protected int mCG_Z_fp8 = 0;
	public int mCG_X = 0;
	public int mCG_Y = 0;
	public int mCG_Z = 0;
	/** forward speed */
	protected int mSpeed_fp8 = 0;
	public int mBoostReserve = 0;
	protected int mBoostFrameCounter = 0;
	/** World x speed */
	protected int mSpeedX_fp8 = 0;
	/** World y speed */
	protected int mSpeedY_fp8 = 0;
	/** World z speed */
	protected int mSpeedZ_fp8 = 0;
	protected int mRotationSpeedX = 0;
	protected int mRotationSpeedY = 0;
	protected int mRotationSpeedZ = 0;
	/** Damage count. if damage count greater then DAMAGE_LIMIT, car is broken */
	public int mDamage = 0;
	/** true if need update damage indicator */
	public boolean mIsUpdateDamage = false;
	/** relative to base point */
	public int mFrontLeftWheelBottom = WHEEL_BOTTOM_Z;
	/** relative to base point */
	public int mFrontRightWheelBottom = WHEEL_BOTTOM_Z;
	/** relative to base point */
	public int mRearLeftWheelBottom = WHEEL_BOTTOM_Z;
	/** relative to base point */
	public int mRearRightWheelBottom = WHEEL_BOTTOM_Z;
	// wheel accelerations
	protected int mFrontLeftWheelAcceleration_fp = 0;;
	protected int mFrontRightWheelAcceleration_fp = 0;
	protected int mRearLeftWheelAcceleration_fp = 0;
	protected int mRearRightWheelAcceleration_fp = 0;
	// wheel to road distances
	protected int mRoadToFrontLeftWheelDistance_fp = 0;
	protected int mRoadToFrontRightWheelDistance_fp = 0;
	protected int mRoadToRearLeftWheelDistance_fp = 0;
	protected int mRoadToRearRightWheelDistance_fp = 0;
	/** to interpolate for smoothing */
	protected int mRoadToCarAngleDifference = 0;
	/** temporary value to calculate wheel position */
	private static int sDistanceToRoadEdge = 0;
	// interaction data
	/**
	 * Car to car interaction delay
	 */
	private static int sInteractionDelay = 0;
	/** distance between cars
	 * positive value: player behind opponent
	 * negative value: opponent behind player
	 */
	public static int sDistance = 0;
	/** 0 - if player not on road<br>
	 * > 0 - if opponent to right from player<br>
	 * < 0 - if opponent to left from player
	 */
	protected static int sDistanceX = 0;
	protected static int sCarToCarXacceleration_fp8 = 0;
	protected static int sCarToCarYacceleration_fp8 = 0;
	protected static int sCarToCarZacceleration_fp8 = 0;
	//
	protected static final short[] mSinToCosTable_fp8 = new short[256 + 1];
	static {
		for( int i = 0; i < mSinToCosTable_fp8.length; i++ ) {
			final double sin2 = ((double)(i * i)) / ((double)(256 * 256));
			mSinToCosTable_fp8[i] = (short)(256 * Math.sqrt( 1.0 - sin2 ));
		}
	}
	//
	//@Override
	public String toString() {
		return mPieceIndex + "[" + mPieceCoordIndex + "], x=" + (mCG_X) + ", y=" + (mCG_Y) + ", z=" + (mCG_Z) +
			", sx=" + mSpeedX_fp8 + ", sy=" + mSpeedY_fp8 + ", sz=" + mSpeedZ_fp8 +
			", ax=" + x_angle + ", ay=" + y_angle + ", az=" + z_angle +
			", fla=" + (mFrontLeftWheelAcceleration_fp >> (Camera.PRECISION_LOG2-3)) +
			", fra=" + (mFrontRightWheelAcceleration_fp >> (Camera.PRECISION_LOG2-3)) +
			", rla=" + (mRearLeftWheelAcceleration_fp >> (Camera.PRECISION_LOG2-3)) +
			", rra=" + (mRearRightWheelAcceleration_fp >> (Camera.PRECISION_LOG2-3));
	}
	protected void setPosition(int x, int y, int z) {
		mCG_X = x;
		mCG_X_fp8 = x << 8;
		mCG_Y = y;
		mCG_Y_fp8 = y << 8;
		mCG_Z = z;
		mCG_Z_fp8 = z << 8;
	}
	/** @return speed to display in km/h */
	public int getSpeedKmH() {// convert sm/period to km/h: 1000(ms)/period(times)/100(sm/m)*3600(s/h)/1000(m/km)
		return (mSpeed_fp8 * 36 / (int)GameView.REFRESH_PERIOD) >> 8;
	}
	/**
	 * Finds current track coordinate index
	 * @param track
	 * @return true if lap number was changed
	 */
	private boolean switchRoadIndex(final TrackPiece[] pieces, final int numTrackCoords)
	{
		boolean is_lap_changed = false;
		int current_index = mCoordIndex;
		boolean is_next;
		do {
			int next_offset = mPieceCoordIndex + 2;
			// find next section
			is_next = Track.isRoadSegmentBehind( pieces[mPieceIndex], next_offset, mCG_X, mCG_Y );
			if( is_next ) {
				if( next_offset >= pieces[mPieceIndex].mCoords.length - 2 ) {
					next_offset = 0;
					if( ++mPieceIndex >= pieces.length ) {
						mPieceIndex -= pieces.length;
					}
					//mPiece = track.mPieces[mPieceIndex];
					if( mPieceIndex == 0 ) {
						mLap++;
						is_lap_changed = true;
					}
				}
				mPieceCoordIndex = next_offset;
				mCoordIndex += 2;
				if( mCoordIndex >= numTrackCoords ) mCoordIndex -= numTrackCoords;
				mTravelDistance++;
			}
		}
		while( is_next );
		if( current_index == mCoordIndex ) {
			// find previous section
			while( Track.isRoadSegmentBefore( pieces[mPieceIndex], mPieceCoordIndex, mCG_X, mCG_Y ) ) {
				mTravelDistance--;
				mPieceCoordIndex -= 2;
				if( mPieceCoordIndex < 0 ) {
					if( mPieceIndex == 0 ) {
						mLap--;
						is_lap_changed = true;
					}
					if( --mPieceIndex < 0 ) {
						mPieceIndex += pieces.length;
					}
					//mPiece = track.mPieces[mPieceIndex];
					mPieceCoordIndex = pieces[mPieceIndex].mCoords.length - 4;
				}
				mCoordIndex -= 2;
				if( mCoordIndex < 0 ) mCoordIndex += numTrackCoords;
			}
		}
		return is_lap_changed;
	}
	/**
	 * Switches road index, sets start race time, calculates best lap time and race time.
	 * Calculates the matrix.
	 * Sets if on road flag.
	 * @param track
	 * @param curr_time
	 * @return 1 - if lap number > 1, 2 - if race finished, 0 - otherwise
	 */
	public int move(final Track track, final long curr_time) {
		int ret = 0;
		if( switchRoadIndex( track.mPieces, track.mNumTrackCoords ) ) {
			//System.out.println("piece = " + mPieceIndex + ", shift = " + mPieceCoordIndex + ", pos = " + mCoordIndex);
			if( mLap > 1 ) {
				ret++;
				final int lap_time = (int)(curr_time - mStartLapTime);
				if( mBestLapTime > lap_time ) {
					mBestLapTime = lap_time;
					ret++;
				}
			}
			mStartLapTime = curr_time;
			if( mLap == 1 )
				mStartRaceTime = curr_time;
			if( mLap > GameView.MAX_LAPS )
				mRaceTime = (int)(curr_time - mStartRaceTime);
		}
		mIsOnRoad = Track.isOnRoad( track.mPieces[mPieceIndex], mPieceCoordIndex, mCG_X, mCG_Y );
		calculateZXYmatrix();
		return ret;
	}

	public void calculateCameraCoordinates() {
		final int mr = 3;
		final int[] m = mMatrix;// for speed optimization
		//
		mX = mCG_X + ((LOOK_Y * m[1 * mr + 0] + LOOK_Z * m[2 * mr + 0]) >> Camera.PRECISION_LOG2);
		mY = mCG_Y + ((LOOK_Y * m[1 * mr + 1] + LOOK_Z * m[2 * mr + 1]) >> Camera.PRECISION_LOG2);
		mZ = mCG_Z + ((LOOK_Y * m[1 * mr + 2] + LOOK_Z * m[2 * mr + 2]) >> Camera.PRECISION_LOG2);
	}

	/**
	 * Calculate height for given point
	 * @param track - current track
	 * @param x - x coordinate for point
	 * @param y - y coordinate for point
	 * @param isCorrectWheelHeight - true if need account edge falling.
	 *     false: just interpolation.
	 * @return height * PRECISION. 0 if off road
	 */
	public final int interpolateHeight(final Track track, int x, int y, boolean isCorrectWheelHeight) {
		int piece = mPieceIndex;
		int next_offset = mPieceCoordIndex;
		int previous_offset = next_offset;
		// find next section
		boolean is_next = true;
		//boolean is_previous = true;
		while( is_next ) {
			is_next = Track.isRoadSegmentBehind( track.mPieces[piece], next_offset, x, y );
			if( is_next ) {
				//is_previous = false;
				previous_offset = next_offset;
				next_offset += 2;
				if( next_offset >= track.mPieces[piece].mCoords.length ) {
					next_offset = 2;
					previous_offset = 0;
					if( ++piece >= track.mPieces.length ) piece -= track.mPieces.length;
				}
			}
		}
		// find previous section
		if( previous_offset == next_offset ) {// is_previous
			while( Track.isRoadSegmentBefore( track.mPieces[piece], previous_offset, x, y ) ) {
				next_offset = previous_offset;
				previous_offset -= 2;
				if( previous_offset < 0 ) {
					if( --piece < 0 ) piece += track.mPieces.length;
					next_offset = track.mPieces[piece].mCoords.length - 2;
					previous_offset = next_offset - 2;
				}
			}
		}
		// if on segment, increase next segment number to prevent incorrect check on road state
		if( previous_offset == next_offset ) {
			next_offset += 2;
		}
		//
		final Coord3D near_left = track.mPieces[piece].mCoords[previous_offset++];
		final Coord3D near_right = track.mPieces[piece].mCoords[previous_offset];
		final Coord3D far_left = track.mPieces[piece].mCoords[next_offset++];
		final Coord3D far_right = track.mPieces[piece].mCoords[next_offset];
		// y vector
		final int dx_ver = far_left.mX - near_left.mX;
		final int dy_ver = far_left.mY - near_left.mY;
		// x vector
		final int dx_hor = near_right.mX - near_left.mX;
		final int dy_hor = near_right.mY - near_left.mY;
		//
		final int dx_left = x - near_left.mX;
		final int dy_left = near_left.mY - y;
		//
		sDistanceToRoadEdge = 0;
		// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
		// orientation > 0 : left
		// orientation < 0 : right
		// orientation == 0 : on
		final int off_left = dx_ver * dy_left + dy_ver * dx_left;
		if( off_left < 0 ) {// off left
			// get distance from and position at left edge
			// calculate perpendicular cross point from left and top edge
			// method is similar to that used when texture mapping
			// bottom edge - calculate surface x position in %
			final int denominator = (dx_ver * dy_hor) - (dy_ver * dx_hor);
			if( denominator != 0 ) {
				//final int ROAD_WIDTH = 380;
				//final int road_width = (int)Math.sqrt( (double)(dx_hor * dx_hor + dy_hor * dy_hor) );
				//sDistanceToRoadEdge = ( (off_left * road_width) / denominator );
				sDistanceToRoadEdge = ( (off_left * track.mRoadWidth) / denominator );
			}
		} else {
			// y vector
			final int dx_right = far_right.mX - near_right.mX;
			final int dy_right = far_right.mY - near_right.mY;
			final int off_right = dx_right * (near_right.mY - y) + dy_right * (x - near_right.mX);
			if( off_right > 0 ) {// off right
				// get distance from and position at left edge
				// calculate perpendicular cross point from left and top edge
				// method is similar to that used when texture mapping
				// bottom edge - calculate surface x position in %
				final int denominator = (dy_right * dx_hor) - (dx_right * dy_hor);
				if( denominator != 0 ) {
					//final int road_width = (int)Math.sqrt( (double)(dx_hor * dx_hor + dy_hor * dy_hor) );
					//sDistanceToRoadEdge = ( (off_right * road_width) / denominator );
					sDistanceToRoadEdge = ( (off_right * track.mRoadWidth) / denominator );
				}
			}
		}
		// calculate perpendicular cross point from left and top edge
		// method is similar to that used when texture mapping
		final int denominator = (dy_ver * dx_hor) - (dx_ver * dy_hor);
		final int FIXED_POINT_LOG2 = 8;// PRECISION_LOG2 is too much, 10 max
		int z_fp = near_left.mZ << FIXED_POINT_LOG2;
		if( denominator != 0 ) {
			// bottom edge - calculate surface x position in %
			final int sx = (off_left << FIXED_POINT_LOG2) / denominator;
			// left edge - calculate surface y position in %
			final int u = dx_left * dy_hor + dy_left * dx_hor;
			final int sy = (u << FIXED_POINT_LOG2) / denominator;// sy has reverse sign
			// calculate z
			// first do x interpolation
			z_fp += (sx * (near_right.mZ - near_left.mZ));// z = znear
			final int zfar = (far_left.mZ << FIXED_POINT_LOG2) + sx * (far_right.mZ - far_left.mZ);
			// now do y interpolation. sy has reverse sign
			z_fp += (sy * (z_fp - zfar)) >> FIXED_POINT_LOG2;
		}
		z_fp <<= (Camera.PRECISION_LOG2 - FIXED_POINT_LOG2);
		if( isCorrectWheelHeight ) {
			if( mIsOnRoad ) {
				if( sDistanceToRoadEdge > 0 ) {// to fall gradually
					z_fp -= (WHEEL_RADIUS << Camera.PRECISION_LOG2);
					z_fp -= (sDistanceToRoadEdge * (WHEEL_RADIUS << Camera.PRECISION_LOG2)) / WHEEL_RIGHT_X;
				}
				return z_fp;
			}
			if( sDistanceToRoadEdge > 0 )
				return 0;// off road
		}
		return z_fp;// wall
	}
	private int getWheelAcceleration(int old_wheel_road_distance_fp, int wheel_road_distance_fp) {
		if( wheel_road_distance_fp > 0) {
			int wheel_height_difference_fp = wheel_road_distance_fp - old_wheel_road_distance_fp;
			if( wheel_height_difference_fp > (int)((981 * GameView.REFRESH_PERIOD / 1000 / 2 / 2) << Camera.PRECISION_LOG2) ) {// damage counter, 0.5g = 1 point
				mDamage += wheel_height_difference_fp / (int)((981 * GameView.REFRESH_PERIOD / 1000 / 2 / 2) << Camera.PRECISION_LOG2);
				if( mDamage > DAMAGE_LIMIT ) mDamage = DAMAGE_LIMIT;
				mIsUpdateDamage = true;
			}
			wheel_height_difference_fp = Math.max( wheel_height_difference_fp, (int)((-981 * GameView.REFRESH_PERIOD / 1000 / 6) << Camera.PRECISION_LOG2) );// 1/12g
			wheel_height_difference_fp = Math.min( wheel_height_difference_fp, (int)((981 * GameView.REFRESH_PERIOD / 1000) << Camera.PRECISION_LOG2) );// 2g
			final int damper_fp8 = (int)(256 * 1000 / GameView.REFRESH_PERIOD / 10);// 10 sm/s
			int acceleration_fp = ((wheel_height_difference_fp * damper_fp8) >> 8) + wheel_road_distance_fp;
			if( acceleration_fp > 0 ) {
				acceleration_fp /= (SPRINT_TRAVEL << (8 - 1)) / -GRAVITY_ACCELERATION_FP8;
				// 10g maximum
				//acceleration_fp = Math.min( acceleration_fp, (-10 * GRAVITY_ACCELERATION_FP8) << (GameView.PRECISION_LOG2 - 8) );
				return acceleration_fp;
			}
		}
		return 0;
	}
	/**
	 * Difference wheel-road is speed
	 * Acceleration is (current speed - previous speed)
	 * Also sets mIsUpdateDamage, mIsTouchingRoad, mIsOffRoad, mIsScratchSide
	 * @param matrix
	 * @param track
	 * @return average acceleration in fp8 format
	 */
	protected int calculateWheelsStroke(final Track track) {
		final int mr = 3;
		final int[] matrix = mMatrix;
		mIsUpdateDamage = false;
		// to account track slope. to avoid calculating angle, assume, car x angle ~ track slope.
		//final int slope_factor_fp = (1 * GameView.PRECISION - cos_x) * WHEEL_RADIUS;
		//--------------------------------------------------------------------------
		// convert local to world coordinates
		// we need only x and y coordinates.
		final int lxm0 = WHEEL_LEFT_X * matrix[0 * mr + 0];
		final int lxm1 = WHEEL_LEFT_X * matrix[0 * mr + 1];
		final int rxm0 = WHEEL_RIGHT_X * matrix[0 * mr + 0];
		final int rxm1 = WHEEL_RIGHT_X * matrix[0 * mr + 1];
		final int fym0 = WHEEL_FRONT_Y * matrix[1 * mr + 0];
		final int fym1 = WHEEL_FRONT_Y * matrix[1 * mr + 1];
		final int rym0 = WHEEL_REAR_Y * matrix[1 * mr + 0];
		final int rym1 = WHEEL_REAR_Y * matrix[1 * mr + 1];
		final int bzm0_x = (mCG_X_fp8 << (Camera.PRECISION_LOG2 - 8)) + WHEEL_BOTTOM_Z * matrix[2 * mr + 0];
		final int bzm1_y = (mCG_Y_fp8 << (Camera.PRECISION_LOG2 - 8)) + WHEEL_BOTTOM_Z * matrix[2 * mr + 1];
		final int front_left_x  = ((lxm0 + fym0 + bzm0_x) >> Camera.PRECISION_LOG2);
		final int front_left_y  = ((lxm1 + fym1 + bzm1_y) >> Camera.PRECISION_LOG2);
		final int front_right_x = ((rxm0 + fym0 + bzm0_x) >> Camera.PRECISION_LOG2);
		final int front_right_y = ((rxm1 + fym1 + bzm1_y) >> Camera.PRECISION_LOG2);
		final int rear_left_x   = ((lxm0 + rym0 + bzm0_x) >> Camera.PRECISION_LOG2);
		final int rear_left_y   = ((lxm1 + rym1 + bzm1_y) >> Camera.PRECISION_LOG2);
		final int rear_right_x  = ((rxm0 + rym0 + bzm0_x) >> Camera.PRECISION_LOG2);
		final int rear_right_y  = ((rxm1 + rym1 + bzm1_y) >> Camera.PRECISION_LOG2);
		final int lxm2 = WHEEL_LEFT_X * matrix[0 * mr + 2];
		final int rxm2 = WHEEL_RIGHT_X * matrix[0 * mr + 2];
		final int fym2 = WHEEL_FRONT_Y * matrix[1 * mr + 2];
		final int rym2 = WHEEL_REAR_Y * matrix[1 * mr + 2];
		final int bzm2_car_z_fp = WHEEL_BOTTOM_Z * matrix[2 * mr + 2] + (mCG_Z_fp8 << (Camera.PRECISION_LOG2 - 8));// + slope_factor_fp;
		// ---- front left wheel
		int height_fp = interpolateHeight( track, front_left_x, front_left_y, true );
		boolean is_left_off_road = (sDistanceToRoadEdge > 0);
		//
		int road_to_wheel = height_fp - (lxm2 + fym2 + bzm2_car_z_fp);
		mFrontLeftWheelAcceleration_fp = getWheelAcceleration( mRoadToFrontLeftWheelDistance_fp, road_to_wheel );
		mRoadToFrontLeftWheelDistance_fp = road_to_wheel;
		// ---- front right wheel
		height_fp = interpolateHeight( track, front_right_x, front_right_y, true );
		boolean is_right_off_road = (sDistanceToRoadEdge > 0);
		//
		road_to_wheel = height_fp - (rxm2 + fym2 + bzm2_car_z_fp);
		mFrontRightWheelAcceleration_fp = getWheelAcceleration( mRoadToFrontRightWheelDistance_fp, road_to_wheel );
		mRoadToFrontRightWheelDistance_fp = road_to_wheel;
		// ---- rear left wheel
		height_fp = interpolateHeight( track, rear_left_x, rear_left_y, true );
		is_left_off_road &= (sDistanceToRoadEdge > 0);
		//
		road_to_wheel = height_fp - (lxm2 + rym2 + bzm2_car_z_fp);
		mRearLeftWheelAcceleration_fp = getWheelAcceleration( mRoadToRearLeftWheelDistance_fp, road_to_wheel );
		mRoadToRearLeftWheelDistance_fp = road_to_wheel;
		// ---- rear right wheel
		height_fp = interpolateHeight( track, rear_right_x, rear_right_y, true );
		is_right_off_road &= (sDistanceToRoadEdge > 0);
		//
		road_to_wheel = height_fp - (rxm2 + rym2 + bzm2_car_z_fp);
		mRearRightWheelAcceleration_fp = getWheelAcceleration( mRoadToRearRightWheelDistance_fp, road_to_wheel );
		mRoadToRearRightWheelDistance_fp = road_to_wheel;
		//--------------------------------------------------------------------------
		mIsTouchingRoad =
				(mRoadToFrontLeftWheelDistance_fp > 0) |
				(mRoadToFrontRightWheelDistance_fp > 0) |
				(mRoadToRearLeftWheelDistance_fp > 0) |
				(mRoadToRearRightWheelDistance_fp > 0);
		//--------------------------------------------------------------------------
		mIsOffRoad = (is_left_off_road & is_right_off_road);
		final boolean is_scratch_side = mIsTouchingRoad & (is_left_off_road | is_right_off_road) & (! mIsOffRoad);
		// sounds
		if( is_scratch_side & ! mIsScratchSide ) {
			SoundManager.playLoop( SoundManager.SCRATCH, is_left_off_road ? 256 : 0, is_right_off_road ? 256 : 0 );
		}
		if( ! is_scratch_side & mIsScratchSide ) {
			SoundManager.stop( SoundManager.SCRATCH );
		}
		//
		mIsScratchSide = is_scratch_side;
		//--------------------------------------------------------------------------
		final int average_acceleration_fp8 =
				((mFrontLeftWheelAcceleration_fp + mFrontRightWheelAcceleration_fp +
				mRearLeftWheelAcceleration_fp + mRearRightWheelAcceleration_fp) >> (2 + Camera.PRECISION_LOG2 - 8));
		return average_acceleration_fp8;
	}
	/** Reset interaction variables */
	public static void resetInteraction() {
		sDistance = 0;
		sInteractionDelay = 0;
	}
	/**
	 * Calculates distance along a track and lateral distance between cars
	 * @param track
	 * @param car
	 * @param opponent
	 */
	public static void interaction(final Track track, final Car car, final Car opponent) {
		// distance between cars along track
		int difference = opponent.mCoordIndex - car.mCoordIndex;
		int distance = Math.abs( difference );
		if( (distance << 1) > track.mNumTrackCoords ) {
			distance -= track.mNumTrackCoords;
			if( difference < 0 ) distance = -distance;
			difference = distance;
		}
		sDistance = difference;
		sDistanceX = 0;
		if( sInteractionDelay > 0 ) {
			sInteractionDelay--;
		}
		if( car.mIsOnRoad ) {
			final int mr = 3;
			final int[] m = car.mMatrix;
			// calculate car to car distance
			final int dx = opponent.mCG_X - car.mCG_X;
			final int dy = opponent.mCG_Y - car.mCG_Y;
			final int dz = opponent.mCG_Z - car.mCG_Z;
			final int rdx = dx * m[0 * mr + 0] + dy * m[0 * mr + 1] + dz * m[0 * mr + 2];
			final int rdy = dx * m[1 * mr + 0] + dy * m[1 * mr + 1] + dz * m[1 * mr + 2];
			final int rdz = dx * m[2 * mr + 0] + dy * m[2 * mr + 1] + dz * m[2 * mr + 2];
			sDistanceX = rdx >> Camera.PRECISION_LOG2;
			if( Math.abs( rdx ) < (WIDTH << Camera.PRECISION_LOG2) &&
				Math.abs( rdy ) < (LENGTH << Camera.PRECISION_LOG2) &&
				Math.abs( rdz ) < (HEIGHT << Camera.PRECISION_LOG2) ) {
				sCarToCarXacceleration_fp8 = (opponent.mSpeedX_fp8 - car.mSpeedX_fp8);// << 1;
				sCarToCarYacceleration_fp8 = (opponent.mSpeedY_fp8 - car.mSpeedY_fp8);// << 1;
				sCarToCarZacceleration_fp8 = (opponent.mSpeedZ_fp8 - car.mSpeedZ_fp8);
				//
				if( sInteractionDelay <= 0 ) {
					SoundManager.play( SoundManager.HIT_CAR, rdx <= 0 ? 256 : 0, rdx >= 0 ? 256 : 0 );
					sInteractionDelay = (int)(1 * 1000 / GameView.REFRESH_PERIOD);
				}
				return;
			}
		}
		sCarToCarXacceleration_fp8 = 0;
		sCarToCarYacceleration_fp8 = 0;
		sCarToCarZacceleration_fp8 = 0;
	}
}