
import java.io.DataInputStream;
import java.io.IOException;
import javax.microedition.lcdui.Graphics;

/**
 * Mask font to be independent from a device
 */
public class MyFont {
	private int[][] mData = null;
	public int mSize = 0;
	public int mWidth = 0;
	public int mBaseLine = 0;
	//
	public static MyFont createFont(int id) {
		DataInputStream dis = null;
		try {
			int k = 1;
			if( (id % 7) == 0 ) {
				k = id / 7;
				id = 75;
			}
			else if( (id % 8) == 0 ) {
				k = id >> 3;
				id = 85;
			}
			else id = 74;
			final int scale_mask = (1 << k) - 1;
			dis = new DataInputStream( Support.class.getResourceAsStream("/font" + id) );
			final MyFont font = new MyFont();
			final int size = (int)dis.readByte();
			final int check_mask = 1 << (size - 1);
			font.mSize = k * size;
			final int width = (int)dis.readByte();
			font.mWidth = k * width;
			font.mBaseLine = k * (int)dis.readByte();
			font.mData = new int[16][];
			for( int i = 0; i < 16; i++ ) {
				final int block_size = dis.readInt();
				final int scaled_block_size = ((1 + width * k) * block_size) / (width + 1);
				final int[] block = new int[scaled_block_size];
				//
				for( int d = 0; d < scaled_block_size; ) {
					block[d++] = (int)dis.readChar();// char code
					for( int s = 0; s < width; s++ ) {
						int mask = (int)dis.readUnsignedByte();// bit mask
						if( k > 1 ) {
							int result = 0;
							for( int n = 0; n < size; n++ ) {
								result <<= k;
								if( (mask & check_mask) != 0 ) {
									result |= scale_mask;
								}
								mask <<= 1;
							}
							mask = result;
							result = k;
							while( --result > 0 ) {
								block[d++] = mask;
							}
						}
						block[d++] = mask;
					}
				}
				font.mData[i] = block;
			}
			return font;
		} catch( Exception e ) {
			System.out.println( e.getMessage() );
		} finally {
			if( dis != null ) try { dis.close(); } catch (IOException ex) {}
		}
		return null;
	}
	public int drawChar(final Graphics g, char c, int x, int y) {
		if( c < ' ' ) return 0;
		int[] block = mData[c & 0xf];
		for( int i = 0; i < block.length; i += mWidth ) {
			if( block[i++] == c ) {
				final int x_end = x + mWidth;
				final int y_end = y + mSize;
				for( int Xshift = x; Xshift < x_end; Xshift++ ) {
					int mask = block[i++];
					int dY = 0;
					for( int Yshift = y; Yshift <= y_end; Yshift++ ) {
						if( (mask & 1) != 0 ) dY++;
						else if( dY != 0 ) {
							g.fillRect( Xshift, Yshift - dY, 1, dY );
							dY = 0;
						}
						mask >>= 1;
					}
					/*if( dY != 0 ) {// font size must be less then 32
						g.fillRect( Xshift, y_end - dY + 1, 1, dY );
					}*/
				}
				return mWidth;
			}
		}
		//System.out.println("dont found font for char: " + c);
        return 0;
    }
}
