public class PlayerCar extends Car {
	// crane data
	public static final int CHAIN_RELEASE_TIME = 1000 / (int)GameView.REFRESH_PERIOD / 2;// 500 ms
	public int mChainReleaseCounter = 0;
	public boolean mIsOnChains = false;
	private int mCraneX_fp8 = 0;
	private int mCraneY_fp8 = 0;
	private int mCraneZ_fp8 = 0;
	private int mTargetX_fp8 = 0;
	private int mTargetY_fp8 = 0;
	/** 50 sm above high point */
	private int mTargetZup_fp8 = 0;
	/** 10 sm above middle point */
	private int mTargetZdown_fp8 = 0;
	//
	private int mEngineRevs = 0;
	/**
	 * Calculate speed of the engine revolutions
	 * @param isAccelerating - true if accelerating or braking
	 * @return engine revolutions speed change
	 */
	private int getEngineRevsChange(final boolean isAccelerating) {
		if( mEngineRevs < 192 ) {// if engine revs are low then increase them slowly
			return 100 * (int)GameView.REFRESH_PERIOD / 1000;// 100 revs/sec
		}
		// revs change depending on current engine revs
		int c = 22 - (mEngineRevs >> 3);// 22 revs/sec is minimal revs speed
		if( mIsTouchingRoad ) {
			c += Math.abs( mSpeed_fp8 >> 7 );
			if( c < -12800 * (int)GameView.REFRESH_PERIOD / 1000 )
				return -12800 * (int)GameView.REFRESH_PERIOD / 1000;// 12800 revs/sec
			return c;
		}
		if( isAccelerating ) {
			c += 28800 * (int)GameView.REFRESH_PERIOD / 1000;// 28800 revs/sec
		}
		if( c >= 12800 * (int)GameView.REFRESH_PERIOD / 1000 ) {
			return 12800 * (int)GameView.REFRESH_PERIOD / 1000;// 12800 revs/sec
		}
		if( c < -1600 * (int)GameView.REFRESH_PERIOD / 1000 ) {
			return  -1600 * (int)GameView.REFRESH_PERIOD / 1000;// 1600 revs/sec
		}
		return c;
	}
	/**
	 * Dependent from car dimensions
	 * Calculate: forward speed
	 * @param matrix
	 * @param track
	 * @param key
	 */
	public void carBehaviour(final Track track, int key) {
		final int mr = 3;
		final int[] m = mMatrix;
		//--------------------------------------------------------------------------
		// calculate forward speed
		mSpeed_fp8 = (mSpeedX_fp8 * m[1 * mr + 0] + mSpeedY_fp8 * m[1 * mr + 1] + mSpeedZ_fp8 * m[1 * mr + 2]) >> (Camera.PRECISION_LOG2);
		//--------------------------------------------------------------------------
		int x_rotation_acceleration = 0;
		int y_rotation_acceleration = 0;
		int z_rotation_acceleration = 0;
		//--------------------------------------------------------------------------
		// calculate accelerations
		int x_acceleration_fp8 = 0;
		int y_acceleration_fp8 = 0;
		int z_acceleration_fp8 = 0;
		//--------------------------------------------------------------------------
		boolean is_boost = false;
		if( (mSpeed_fp8 < MAX_SPEED_FP8) && (mDamage < DAMAGE_LIMIT) &&
			(mIsAccelerating || ((key & (GameView.KEY_UP | GameView.KEY_DOWN)) != 0)) )
		{
			if( (key & GameView.KEY_FIRE) != 0 ) {
				if( mBoostReserve > 0 ) {
					if( ++mBoostFrameCounter >= (1000 / (int)GameView.REFRESH_PERIOD) ) {// 1.0 сек
						mBoostFrameCounter = 0;
						mBoostReserve--;
					}
					is_boost = true;
				}
			}
			mIsAccelerating = (key & GameView.KEY_DOWN) == 0;// backward
		}// if accelerating or decelerating
		mIsBoostActivated = is_boost;
		//
		//System.out.println("player " + mPieceIndex + ", speed = " + (mSpeed_fp8 >> 8) + ", key = " + Integer.toHexString( key ) );
		// calculate engine revolutions
		mEngineRevs += getEngineRevsChange( (key & (GameView.KEY_DOWN | GameView.KEY_UP | GameView.KEY_FIRE)) != 0 );
		mEngineRevs = Math.max( mEngineRevs, 0 );
		SoundManager.engine( mEngineRevs );
		if( mIsOnChains ) return;
		//--------------------------------------------------------------------------
		//final int slope_factor_fp = WHEEL_RADIUS * track.getSectionSlopeCos( mCoordIndex );
		final int average_acceleration_fp8 = calculateWheelsStroke( track );
		if( mIsTouchingRoad ) {
			/*
			cos = sqrt( 1*1 - sin*sin ) = mSinToCos[sin]
			angles in range -90 ... +90 degrees
			z_inclination_to_road is zero because road exists in X and Y planes only
			calculate ZXY matrix to convert from world to local coordinates
			x = x * cos z * cos y + x * sin z * sin x * sin y + y * cos z * sin x * sin y - y * sin z * cos y + z * cos x * sin y
			y = x * sin z * cos x + y * cos z * cos x - z * sin x
			z = x * sin z * sin x * cos y - x * cos z * sin y + y * cos z * sin x * cos y + y * sin z * sin y + z * cos x * cos y
			we have only z value, so
			x = z * cos x * sin y
			y = - z * sin x = z * (- sin x)
			z = z * cos x * cos y
			To simplify, calculate sin x with reverse sign */
			//--------------------------------------------------------------------------
			//
			final int longtitude_difference_fp8 = (mRoadToRearLeftWheelDistance_fp + mRoadToRearRightWheelDistance_fp -
					mRoadToFrontLeftWheelDistance_fp - mRoadToFrontRightWheelDistance_fp) >> (1 + Camera.PRECISION_LOG2 - 8);
			//
			//final int longtitude_difference_fp8 =
			//	(Math.max( mRoadToRearLeftWheelDistance_fp, mRoadToRearRightWheelDistance_fp ) -
			//	Math.max( mRoadToFrontLeftWheelDistance_fp, mRoadToFrontRightWheelDistance_fp) ) >> (Camera.PRECISION_LOG2 - 8);
			//
			/* Calculate x inclination to road, sin x with reverse sign
			tan x = (df - dr) / length
			but for small angle can use:
			sin x = (df - dr) / length
			and need to limit result to 1 */
			final int surface_neg_sinx_fp8 = Math.min( Math.max( (longtitude_difference_fp8) / WHEEL_FRONT_REAR_DISTANCE, -256 ), 256 );
			// get cosines
			final int surface_cosx_fp8 = (int)mSinToCosTable_fp8[Math.abs( surface_neg_sinx_fp8 )];
			/*
			final int lateral_difference_fp8 = (mRoadToFrontLeftWheelDistance_fp + mRoadToRearLeftWheelDistance_fp -
					mRoadToFrontRightWheelDistance_fp - mRoadToRearRightWheelDistance_fp) >> (1 + Camera.PRECISION_LOG2 - 8);
			*/
			final int lateral_difference_fp8 =
				(Math.max( mRoadToFrontLeftWheelDistance_fp, mRoadToRearLeftWheelDistance_fp) -
				Math.max( mRoadToFrontRightWheelDistance_fp, mRoadToRearRightWheelDistance_fp ) ) >> (Camera.PRECISION_LOG2 - 8);
			//
			/* Calculate y inclination to road
			tan y = (dfl - dfr) / length
			but for small angle can use:
			sin x = (dfl - dfr) / length
			and need to limit result to 1 */
			final int surface_siny_fp8 = Math.min( Math.max( (lateral_difference_fp8) / WHEEL_TREAD, -256 ), 256 );
			// get cosines
			final int surface_cosy_fp8 = (int)mSinToCosTable_fp8[Math.abs( surface_siny_fp8 )];
			//--------------------------------------------------------------------------
			// car_road_collision_x_acceleration
			int car_road_x_acceleration_fp8 = (average_acceleration_fp8 * ((surface_cosx_fp8 * surface_siny_fp8) >> 8)) >> (8);
			// car_road_collision_y_acceleration
			final int car_road_y_acceleration_fp8 = (average_acceleration_fp8 * surface_neg_sinx_fp8) >> (8);
			// car_road_collision_z_acceleration
			final int car_road_z_acceleration_fp8 = (average_acceleration_fp8 * ((surface_cosx_fp8 * surface_cosy_fp8) >> 8)) >> (8);
			// car_road_z_acceleration always positive
			final int double_car_road_z_acceleration_fp8 = (car_road_z_acceleration_fp8 << 1);
			//--------------------------------------------------------------------------
			// x and y accelerations is forward moving
			// x acceleration
			if( Math.abs( car_road_x_acceleration_fp8 ) > double_car_road_z_acceleration_fp8 ) {
				int x_accel = double_car_road_z_acceleration_fp8;
				if( car_road_x_acceleration_fp8 < 0 )
					x_accel = -x_accel;
				car_road_x_acceleration_fp8 = x_accel;
			}
			final int x_speed_fp8 = (mSpeedX_fp8 * m[0 * mr + 0] + mSpeedY_fp8 * m[0 * mr + 1] + mSpeedZ_fp8 * m[0 * mr + 2]) >> (Camera.PRECISION_LOG2);
			car_road_x_acceleration_fp8 -= x_speed_fp8;
			x_acceleration_fp8 += car_road_x_acceleration_fp8;
			y_acceleration_fp8 += car_road_y_acceleration_fp8;
			z_acceleration_fp8 += car_road_z_acceleration_fp8;
			//--------------------------------------------------------------------------
			if( mIsOffRoad ) {//|| (mDamage > DAMAGE_LIMIT && ! mIsScratchSide) ) {// brake if wrecked or touching ground
				x_acceleration_fp8 -= x_speed_fp8 >> 1;
				y_acceleration_fp8 -= mSpeed_fp8 >> 1;
				//z_acceleration_fp8 -= 0;
				x_rotation_acceleration -= mRotationSpeedX >> 4;
				y_rotation_acceleration -= mRotationSpeedY >> 4;
				z_rotation_acceleration -= mRotationSpeedZ >> 1;
			}// if is off road
			// if acceleration > 0.5 g
			if( average_acceleration_fp8 > (int)((981 * GameView.REFRESH_PERIOD / 1000 / 2 / 2 ) << 8) ) {
				SoundManager.play( SoundManager.SMASH, 256, 256 );
			} else {// if acceleration > 0.25 g
				if( average_acceleration_fp8 > (int)((981 * GameView.REFRESH_PERIOD / 1000 / 2 / 4) << 8) )
					SoundManager.play( mIsOffRoad ? SoundManager.GROUNDED : SoundManager.CREAK, 256, 256 );
			}
			//--------------------------------------------------------------------------
			if( mRoadToFrontLeftWheelDistance_fp > 0 || mRoadToFrontRightWheelDistance_fp > 0 ) {// left or right wheel touching the road
				// calculate difference angle
				int look_ahead_piece = mPieceIndex;
				int look_ahead_index = mPieceCoordIndex + 1 * 2;
				if( look_ahead_index >= track.mPieces[look_ahead_piece].mCoords.length - 2 ) {
					look_ahead_index -= mPieceCoordIndex;
					look_ahead_index += 2;
					if( ++look_ahead_piece >= track.mPieces.length ) look_ahead_piece -= track.mPieces.length;
				}
				int angle_difference = Track.getSectionAngle( track.mPieces[look_ahead_piece], look_ahead_index ) - z_angle;
				if( angle_difference > Camera.ANGLE_180 ) angle_difference -= Camera.ANGLE_360;
				if( angle_difference < -Camera.ANGLE_180) angle_difference += Camera.ANGLE_360;
				// smoothing
				//mRoadToCarAngleDifference = ((angle_difference + mRoadToCarAngleDifference) >> 1);
				mRoadToCarAngleDifference += ((angle_difference - mRoadToCarAngleDifference) >> 2);
				// manual or automatic steering
				if( (key & (GameView.KEY_LEFT | GameView.KEY_RIGHT)) != 0 ) {// if player steering
					int steering_z_rotation_acceleration = mSpeed_fp8 >> (8 - 0);
					if( mRoadToFrontLeftWheelDistance_fp > 0 && mRoadToFrontRightWheelDistance_fp > 0 ) {
						// twice if both wheels touch the road
						steering_z_rotation_acceleration += steering_z_rotation_acceleration;
					}
					steering_z_rotation_acceleration *= surface_cosx_fp8;
					steering_z_rotation_acceleration >>= 8;
					if( (key & GameView.KEY_LEFT) != 0 ) {
						steering_z_rotation_acceleration = -steering_z_rotation_acceleration;
					}
					if( (steering_z_rotation_acceleration ^ angle_difference) > 0 ) {// same direction
						// steer helper
						steering_z_rotation_acceleration += (mRoadToCarAngleDifference * Math.abs( mSpeed_fp8 )) >> 17;
					}
					// adjusting.
					//steering_z_rotation_acceleration *= 296;
					//steering_z_rotation_acceleration >>= 8;
					//
					z_rotation_acceleration += steering_z_rotation_acceleration;
					z_rotation_acceleration -= mRotationSpeedZ;
					mRotationSpeedZ += z_rotation_acceleration;
				} else {// if player not steering
					int diff_angle = Math.min( Math.abs( angle_difference ), 2 * Camera.ANGLE_360 / 360 );
					diff_angle *= Math.abs( mSpeed_fp8 );
					diff_angle >>= (8 + 8);
					diff_angle = Math.min( diff_angle, 10 * Camera.ANGLE_360 / 360 );
					diff_angle = Math.max( diff_angle, 1 );
					if( angle_difference < 0 )
						diff_angle = -diff_angle;
					z_angle += diff_angle;
					mRotationSpeedZ = 0;
					//mRotationSpeedZ = diff_angle;
				}// if player not steering
			}// left or right wheel touching the road
			//--------------------------------------------------------------------------
			if( mIsAccelerating || ( (key & GameView.KEY_DOWN) != 0 ) ) {
				// engine acceleration depends from current forward speed
				int engine_acceleration_fp8 = -(mSpeed_fp8 >> 8);//-(Math.abs( mSpeed_fp8 ) >> 7);
				if( mRoadToFrontLeftWheelDistance_fp > 0 ) engine_acceleration_fp8 += ENGINE_ACCELERATION_FP8;
				if( mRoadToFrontRightWheelDistance_fp > 0 ) engine_acceleration_fp8 += ENGINE_ACCELERATION_FP8;
				if( mRoadToRearLeftWheelDistance_fp > 0 ) engine_acceleration_fp8 += ENGINE_ACCELERATION_FP8;
				if( mRoadToRearRightWheelDistance_fp > 0 ) engine_acceleration_fp8 += ENGINE_ACCELERATION_FP8;
				if( mIsBoostActivated ) {
					engine_acceleration_fp8 <<= 1;
				}
				// car accelerate depends from grip
				engine_acceleration_fp8 = Math.min( engine_acceleration_fp8, double_car_road_z_acceleration_fp8 );
				if( (key & GameView.KEY_DOWN) != 0 ) {// backward
					engine_acceleration_fp8 = -engine_acceleration_fp8;
				}
				y_acceleration_fp8 += engine_acceleration_fp8;
				// to pitch while accelerating
				x_rotation_acceleration -= (engine_acceleration_fp8 >> (8 - 4));
			}
			//--------------------------------------------------------------------------
			// ration x/y is WHEEL_FRONT_REAR_DISTANCE / (2 * WHEEL_X) ~= 2
			x_rotation_acceleration += (mRearLeftWheelAcceleration_fp + mRearRightWheelAcceleration_fp -
				mFrontLeftWheelAcceleration_fp - mFrontRightWheelAcceleration_fp) >> (1 + Camera.PRECISION_LOG2 - 5);// 6 max
			//
			y_rotation_acceleration = (mFrontRightWheelAcceleration_fp + mRearRightWheelAcceleration_fp -
				mFrontLeftWheelAcceleration_fp - mRearLeftWheelAcceleration_fp) >> (1 + Camera.PRECISION_LOG2 - 7);// 9 max
		}// if mIsTouchingRoad
		// if on edge or fly do small forward rotation
		if( ! mIsTouchingRoad ) {// 17 deg/s^2
			x_rotation_acceleration = Math.max( x_rotation_acceleration, (int)(17 * Camera.ANGLE_360 / 360 * GameView.REFRESH_PERIOD * GameView.REFRESH_PERIOD / 1000 / 1000) );
		}
		x_rotation_acceleration -= (mRotationSpeedX >> 4);
		y_rotation_acceleration -= (mRotationSpeedY >> 4);
		//--------------------------------------------------------------------------
		// the effect of air resistance and road load
		// Check if slipstream
		if( sDistance > 0 && sDistance < 4 * 2) {// if close to player
			if( Math.abs( sDistanceX ) < WIDTH ) {
				y_acceleration_fp8 += mSpeed_fp8 >> 16;
			}
		}
		y_acceleration_fp8 -= mSpeed_fp8 >> 8;
		int world_x_acceleration_fp8 = (x_acceleration_fp8 * m[0 * mr + 0] + y_acceleration_fp8 * m[1 * mr + 0] + z_acceleration_fp8 * m[2 * mr + 0]) >> (Camera.PRECISION_LOG2);
		int world_y_acceleration_fp8 = (x_acceleration_fp8 * m[0 * mr + 1] + y_acceleration_fp8 * m[1 * mr + 1] + z_acceleration_fp8 * m[2 * mr + 1]) >> (Camera.PRECISION_LOG2);
		int world_z_acceleration_fp8 = (x_acceleration_fp8 * m[0 * mr + 2] + y_acceleration_fp8 * m[1 * mr + 2] + z_acceleration_fp8 * m[2 * mr + 2]) >> (Camera.PRECISION_LOG2);
		world_z_acceleration_fp8 += GRAVITY_ACCELERATION_FP8;
		// speed = curr_speed + acceleration * dt. dt = 1;
		mSpeedX_fp8 += world_x_acceleration_fp8;
		mSpeedY_fp8 += world_y_acceleration_fp8;
		mSpeedZ_fp8 += world_z_acceleration_fp8;
		//
		mSpeedX_fp8 += sCarToCarXacceleration_fp8;
		mSpeedY_fp8 += sCarToCarYacceleration_fp8;
		mSpeedZ_fp8 += sCarToCarZacceleration_fp8;
		//
		mCG_X_fp8 += mSpeedX_fp8;
		mCG_X = mCG_X_fp8 >> 8;
		mCG_Y_fp8 += mSpeedY_fp8;
		mCG_Y = mCG_Y_fp8 >> 8;
		mCG_Z_fp8 += mSpeedZ_fp8;
		mCG_Z = mCG_Z_fp8 >> 8;
		//--------------------------------------------------------------------------
		mFrontLeftWheelBottom = WHEEL_BOTTOM_Z +
			Math.min( Math.max( mRoadToFrontLeftWheelDistance_fp >> Camera.PRECISION_LOG2, 0 ), SPRINT_TRAVEL );

		mFrontRightWheelBottom = WHEEL_BOTTOM_Z +
			Math.min( Math.max( mRoadToFrontRightWheelDistance_fp >> Camera.PRECISION_LOG2, 0 ), SPRINT_TRAVEL );

		mRearLeftWheelBottom = WHEEL_BOTTOM_Z +
				Math.min( Math.max( mRoadToRearLeftWheelDistance_fp >> Camera.PRECISION_LOG2, 0 ), SPRINT_TRAVEL );

		mRearRightWheelBottom = WHEEL_BOTTOM_Z +
			Math.min( Math.max( mRoadToRearRightWheelDistance_fp >> Camera.PRECISION_LOG2, 0 ), SPRINT_TRAVEL );
		//--------------------------------------------------------------------------
		setZangle( z_angle + mRotationSpeedZ );// z-1 - left
		//
		mRotationSpeedY += y_rotation_acceleration;
		mRotationSpeedX += x_rotation_acceleration;
		int angle = x_angle + mRotationSpeedX;
		if( angle > 65 * Camera.ANGLE_360 / 360 ) {
			angle = 65 * Camera.ANGLE_360 / 360;
			mRotationSpeedX = 0;
		} else if( angle < -65 * Camera.ANGLE_360 / 360 ) {
			angle = -65 * Camera.ANGLE_360 / 360;
			mRotationSpeedX = 0;
		}
		setXangle( angle );// x-1 - forward UP
		//
		angle = y_angle + mRotationSpeedY;
		if( angle > 65 * Camera.ANGLE_360 / 360 ) {
			angle = 65 * Camera.ANGLE_360 / 360;
			mRotationSpeedY = 0;
		} else if( angle < -65 * Camera.ANGLE_360 / 360 ) {
			angle = -65 * Camera.ANGLE_360 / 360;
			mRotationSpeedY = 0;
		}
		setYangle( angle );// y-1 - clockwise
	}
	public void deccelerateEngine() {
		// calculate engine revolutions
		mEngineRevs += -1600 * (int)GameView.REFRESH_PERIOD / 1000;// 1600 revs/sec;
		mEngineRevs = Math.max( mEngineRevs, 0 );
		SoundManager.engine( mEngineRevs );
	}
	public void hookOnCrane(final TrackPiece[] pieces) {
		mIsOnChains = true;
		mChainReleaseCounter = 0;
		mIsTouchingRoad = false;
		mCraneX_fp8 = mCG_X_fp8;
		mCraneY_fp8 = mCG_Y_fp8;
		mCraneZ_fp8 = mCG_Z_fp8;
		final TrackPiece tp = pieces[mPieceIndex];
		setAngles( 0, 0, Track.getSectionAngle( tp, mPieceCoordIndex ) );
		calculateZXYmatrix();
		// calculate target point
		final int mr = 3;
		final int shift = (WIDTH / 2 + CENTER_LINE_MARGIN);
		final Coord3D coord_left = tp.mCoords[mPieceCoordIndex];
		final Coord3D coord_right = tp.mCoords[mPieceCoordIndex + 1];
		final int z = ((coord_left.mZ + coord_right.mZ) >> 1) + CG_Z;
		mTargetZup_fp8 = (50 + z + Math.abs( (coord_left.mZ - coord_right.mZ) >> 1 )) << 8;
		mTargetZdown_fp8 = (CG_Z + Math.max( coord_left.mZ, coord_right.mZ )) << 8;
		mTargetX_fp8 = (coord_left.mX + coord_right.mX) << (8 - 1);
		mTargetY_fp8 = (coord_left.mY + coord_right.mY) << (8 - 1);
		mTargetX_fp8 += (mMatrix[0 * mr + 0] * shift) >> (Camera.PRECISION_LOG2 - 8);
		mTargetY_fp8 += (mMatrix[0 * mr + 1] * shift) >> (Camera.PRECISION_LOG2 - 8);
		// reset car state
		mFrontLeftWheelBottom = WHEEL_BOTTOM_Z;
		mFrontRightWheelBottom = WHEEL_BOTTOM_Z;
		mRearLeftWheelBottom = WHEEL_BOTTOM_Z;
		mRearRightWheelBottom = WHEEL_BOTTOM_Z;
		mRoadToFrontLeftWheelDistance_fp = 0;
		mRoadToFrontRightWheelDistance_fp = 0;
		mRoadToRearLeftWheelDistance_fp = 0;
		mRoadToRearRightWheelDistance_fp = 0;
		mRotationSpeedX = 0;
		mRotationSpeedY = 0;
		mRotationSpeedZ = 0;
		mSpeedX_fp8 = 0;
		mSpeedY_fp8 = 0;
		mSpeedZ_fp8 = 0;
		mSpeed_fp8 = 0;
		mIsAccelerating = false;
	}
	/**
	 * Set initial parameters for player car
	 * @param track
	 */
	public void setOnCrane(final Track track) {
		final int distance_to_start = 4;
		mLap = 0;
		mPieceIndex = track.mPieces.length;
		TrackPiece tp = track.mPieces[--mPieceIndex];
		mPieceCoordIndex = tp.mCoords.length - (distance_to_start + 1) * 2;
		mCoordIndex = track.mNumTrackCoords - distance_to_start * 2;
		/*
		mPieceIndex = 0;
		tp = track.mPieces[mPieceIndex];
		mPieceCoordIndex = tp.mCoords.length - (4 + 1) * 2;
		mCoordIndex = mPieceCoordIndex;
		for( int i = 0; i < mPieceIndex; i++ ) {
			mCoordIndex += track.mPieces[i].mCoords.length;
			mCoordIndex -= 2;
		}
		*/
		mTravelDistance = 0;
		mDamage = 0;
		mStartRaceTime = 0;
		mBestLapTime = 0x7fffffff;
		mRaceTime = 0x7fffffff;
		mBoostReserve = track.mStandardBoost;
		mBoostFrameCounter = 0;
		mIsBoostActivated = false;
		mEngineRevs = 0;
		//
		hookOnCrane( track.mPieces );// to calculate view matrix
		// car sets on ground with shift to right 1024 sm from center
		final int mr = 3;
		final int shift = 1024;// -(WIDTH / 2 + CENTER_LINE_MARGIN)
		final Coord3D coord_left = tp.mCoords[mPieceCoordIndex];
		final Coord3D coord_right = tp.mCoords[mPieceCoordIndex + 1];
		int x = (coord_left.mX + coord_right.mX) >> 1;
		int y = (coord_left.mY + coord_right.mY) >> 1;
		int z = CG_Z;
		x += ((mMatrix[0 * mr + 0] * shift) >> Camera.PRECISION_LOG2);
		y += ((mMatrix[0 * mr + 1] * shift) >> Camera.PRECISION_LOG2);
		//z += (coord_left.mZ + coord_right.mZ) >> 1;
		setPosition( x, y, z );
		// after changing position need set crane coordinates
		mCraneX_fp8 = mCG_X_fp8;
		mCraneY_fp8 = mCG_Y_fp8;
		mCraneZ_fp8 = mCG_Z_fp8;
	}
	private void swing(final int diff) {
		final int CHAIN_LENGTH = 500;// sm
		//
		int acceleration = (int)(Camera.ANGLE_360 * GameView.REFRESH_PERIOD / 360 / 1000);
		if( y_angle >= 0 ) acceleration = -acceleration;
		acceleration += (diff * Camera.ANGLE_360 / (CHAIN_LENGTH * 6 * 30));
		acceleration -= y_angle >> 10;
		//acceleration -= mRotationSpeedY >> 7;
		// angle acceleration is arcsin(dx/CHAIN_LENGTH)
		// for small angles sin a ~= a, so a = dx / CHAIN_LENGTH * 360 / 2pi
		mRotationSpeedY += acceleration;// + (diff * GameView.ANGLE_360 / (CHAIN_LENGTH * 6 * 30));
		setYangle( y_angle + mRotationSpeedY );
		// car coordinates
		int dx_fp8 = (CHAIN_LENGTH * sin_y) >> (Camera.PRECISION_LOG2 - 8);
		// around z. local to world. dy = 0
		//x = x * cos z + y * sin z
		//y = y * cos z - x * sin z
		int x = (mCraneX_fp8 + ((dx_fp8 * cos_z) >> Camera.PRECISION_LOG2)) >> 8;
		int y = (mCraneY_fp8 - ((dx_fp8 * sin_z) >> Camera.PRECISION_LOG2)) >> 8;
		int z = (mCraneZ_fp8 + (CHAIN_LENGTH << 8));
		z -= ((CHAIN_LENGTH * cos_y) >> (Camera.PRECISION_LOG2 - 8));
		z >>= 8;
		setPosition( x, y, z );
	}
	/**
	 * @return true if car was dropped
	 */
	public boolean moveCrane() {
		int dx_fp8 = mTargetX_fp8 - mCraneX_fp8;
		int dy_fp8 = mTargetY_fp8 - mCraneY_fp8;
		if( (Math.abs( dx_fp8 ) < 16 * 256) && (Math.abs( dy_fp8 ) < 16 * 256) ) {// down or drop
			int dz_fp8 = mTargetZdown_fp8 - mCraneZ_fp8;
			if( dz_fp8 < 0 ) {// down, min 100 sm/s
				// down
				dz_fp8 /= (int)(1 * 1000 / GameView.REFRESH_PERIOD);
				dz_fp8 = Math.min( dz_fp8, -(int)(256 * 100 * GameView.REFRESH_PERIOD / 1000) );
				mCraneZ_fp8 += dz_fp8;
				swing( 0 );
				return false;
			}
			// drop
			dx_fp8 = mTargetX_fp8 - mCG_X_fp8;
			dy_fp8 = mTargetY_fp8 - mCG_Y_fp8;
			if( (Math.abs( dx_fp8 ) < 20 * 256) && (Math.abs( dy_fp8 ) < 20 * 256)/* &&
					(Math.abs( y_angle ) < 5 * GameView.ANGLE_360 / 360) */) {
				mIsOnChains = false;
			}
			swing( 0 );
			return ! mIsOnChains;
		}
		// up or move
		int dz_fp8 = mTargetZup_fp8 - mCraneZ_fp8;
		if( dz_fp8 > 0 ) {// up, min 100 sm/s
			dz_fp8 /= (int)(1 * 1000 / GameView.REFRESH_PERIOD);
			dz_fp8 = Math.max( dz_fp8, (int)(256 * 100 * GameView.REFRESH_PERIOD / 1000) );
			mCraneZ_fp8 += dz_fp8;
			//move( 0, 0, dz_fp8 );
			//mCG_Z_fp8 += dz_fp8;
			//mCG_Z = mCG_Z_fp8 >> 8;
			setPosition( mCG_X, mCG_Y, (mCG_Z_fp8 + dz_fp8) >> 8 );
			return false;
		}
		// move
		dx_fp8 /= (int)(1 * 1000 / GameView.REFRESH_PERIOD);
		dy_fp8 /= (int)(1 * 1000 / GameView.REFRESH_PERIOD);
		int diff = (Math.abs(dx_fp8) + Math.abs(dy_fp8)) >> 8;
		diff = Math.min( diff, (int)(1 * 1000 / GameView.REFRESH_PERIOD) );
		// around z. world to local
		//x = x * cos z - y * sin z
		if( (dx_fp8 * cos_z - dy_fp8 * sin_z) > 0 ) {
			diff = -diff;
		}
		mCraneX_fp8 += dx_fp8;
		mCraneY_fp8 += dy_fp8;
		swing( diff );
		return false;
	}
}