
import javax.microedition.lcdui.Graphics;

public class Render {
	/** view distance between viewport and eye */
	public static final int CAMERA_FOV = 32;// cm
	/** 3D to 2D stretch factor */
	public static final int SCALE = (GameView.SCREEN_WIDTH - 16 * 2 * GameView.HPF8 / 256) * (Car.WHEEL_FRONT_Y + CAMERA_FOV) / Car.WHEEL_TREAD;// 16 is pillar width
	/** horizon line position */
	public static final int LOOK_Y = GameView.SCREEN_HALFHEIGHT - 23 * GameView.VPF8 / 256;
	/** distance for detail view */
	private static final int TRACK_ACCURATE_VIEW = 11;// sections
	/** steep angle = z_difference / step */
	private static final int STEEP_Z_DIFFERENCE = 160;
	private static final int OFF_DISTANCE = -((65536>>3) * (65536>>3) * 2);
	private static Section[] sSections = null;
	/** to sort */
	private static Section[] sAuxSections = null;
	/** to save section data */
	private static Section sSectionStore = null;
	private static Section sSectionStoreNext = null;
	/** current track to use colors and to calculate near sections */
	private static Track sTrack = null;
	/** projection mode */
	public static boolean sIsPreview = false;
	/** points to draw polygons. 24 screen points maximum, x coordinates */
	private static final int mXPoints[] = new int[OpponentCar.WHEEL_CIRCLE_POINTS * 2];
	/** points to draw polygons. 24 screen points maximum, y coordinates */
	private static final int mYPoints[] = new int[OpponentCar.WHEEL_CIRCLE_POINTS * 2];
	/** points to draw polygons. 4 screen points maximum, x coordinates */
	private static final int mXPoints4[] = new int[4];// to draw 4 points polygon
	/** points to draw polygons. 4 screen points maximum, y coordinates */
	private static final int mYPoints4[] = new int[4];// to draw 4 points polygon
	/** dimension to store calculated polygon's x coordinates. 22 points */
	private static final int mScreenCoordinatesX[] = new int[18 + 4];
	/** dimension to store calculated polygon's y coordinates. 22 points */
	private static final int mScreenCoordinatesY[] = new int[18 + 4];
	/**
	 * Matrix to transform coordinates, mr = 4<br>
	 * x = [0*mr+0] * x + [0*mr+1] * y + [0*mr+2] * z + [0*mr+3]<br>
	 * y = [1*mr+0] * x + [1*mr+1] * y + [1*mr+2] * z + [1*mr+3]<br>
	 * z = [2*mr+0] * x + [2*mr+1] * y + [2*mr+2] * z + [2*mr+3]<br>
	 */
	private static final int[] mTransformMatrix = new int[3 * 4];
	//
	/** Color to draw the ground */
	private static final int COLOR_GROUND = 0x606040;
	/** Color to draw the sky */
	private static final int COLOR_SKY = 0x4080e0;
    /**
     * Src is the source array that starts at index 0
     * Dest is the (possibly larger) array destination with a possible offset
     * low is the index in dest to start sorting
     * high is the end index in dest to end sorting
     * off is the offset to generate corresponding low, high in src
     */
	private static void mergeSort(Object[] src, Object[] dest,
			int low, int high, int off) {
		int length = high - low;

		// Insertion sort on smallest arrays
		final int INSERTIONSORT_THRESHOLD = 7;
		if( length < INSERTIONSORT_THRESHOLD ) {
			for( int i = low; i < high; i++ ) {
				for( int j = i;
					j > low && ((MyComparable) dest[j - 1]).compareTo( dest[j] ) > 0;
					j-- )
				{
					Object t = dest[j];
					dest[j] = dest[j - 1];
					dest[j - 1] = t;
				}
			}
			return;
		}

		// Recursively sort halves of dest into src
		int destLow  = low;
		int destHigh = high;
		low  += off;
		high += off;
		int mid = (low + high) >>> 1;
		mergeSort( dest, src, low, mid, -off );
		mergeSort( dest, src, mid, high, -off );

		// If list is already sorted, just copy from src to dest.  This is an
		// optimization that results in faster sorts for nearly ordered lists.
		if( ((MyComparable)src[mid - 1]).compareTo( src[mid] ) <= 0) {
			System.arraycopy( src, low, dest, destLow, length );
			return;
		}

		// Merge sorted halves (now in src) into dest
		for( int i = destLow, p = low, q = mid; i < destHigh; i++ ) {
			if( q >= high || p < mid && ((MyComparable)src[p]).compareTo( src[q] ) <= 0 )
				dest[i] = src[p++];
			else
				dest[i] = src[q++];
		}
	}
	/** Uses camera angles */
	public static void drawHorizon(final Graphics g, final Camera camera) {
		// rotate XY. horizon don't depends from Z angle.
		// around x
		//y = y * cos x - z * sin x
		//z = z * cos x + y * sin x
		// around y
		//x = x * cos y + z * sin y
		//z = z * cos y - x * sin y
		//----------
		// x = x * cos y + y * sin x * sin y + z * cos x * sin y
		// y = y * cos x - z * sin x
		// z = - x * sin y + y * sin x * cos y + z * cos x * cos y
		//
		// x1.y1.z1 = -max.max.z
		// x2.y2.z2 =  max.max.z
		//
		final int max_log2 = 16;// 65536;
		final int temp1 = (camera.sin_x << max_log2) >> Camera.PRECISION_LOG2;
		final int temp2 = (camera.mZ * camera.cos_x) >> Camera.PRECISION_LOG2;
		final int temp3 = (temp1 - temp2) * camera.sin_y;
		final int xc = camera.cos_y << max_log2;
		final int xl = (temp3 - xc);
		final int xr = (temp3 + xc);
		final int temp4 = (temp1 - temp2) * camera.cos_y;
		final int xs = camera.sin_y << max_log2;
		final int zl = (temp4 + xs);
		final int zr = (temp4 - xs);
		final int distance = (CAMERA_FOV << Camera.PRECISION_LOG2) + ((camera.cos_x << max_log2) + camera.mZ * camera.sin_x);
		// world to screen
		final int scale = distance / SCALE;
		int x1 = GameView.SCREEN_HALFWIDTH + (xl / scale);
		int y1 = LOOK_Y - (zl / scale);
		int x2 = GameView.SCREEN_HALFWIDTH + (xr / scale);
		int y2 = LOOK_Y - (zr / scale);
		//g.drawLine( x1, z1, x2, z2 );
		g.setColor( COLOR_SKY );
		g.fillRect( 0, 0, GameView.SCREEN_WIDTH, GameView.SCREEN_HEIGHT );
		g.setColor( COLOR_GROUND );
		// check position on the screen
		boolean is_upside_down = false;
		if( x1 > x2 ) {
			if( x1 >= GameView.SCREEN_WIDTH && x2 >= GameView.SCREEN_WIDTH ) return;// only sky
			if( y1 < 0 && y2 < 0 ) return;// only sky
			if( (x1 < 0 && x2 < 0) || (y1 >= GameView.SCREEN_HEIGHT && y2 >= GameView.SCREEN_HEIGHT) ) {// only ground
				g.fillRect( 0, 0, GameView.SCREEN_WIDTH, GameView.SCREEN_HEIGHT );
				return;
			}
			is_upside_down = true;
			int t = x1;
			x1 = x2;
			x2 = t;
			t = y1;
			y1 = y2;
			y2 = t;
		}
		if( ! is_upside_down ) {
			if( y1 >= GameView.SCREEN_HEIGHT && y2 >= GameView.SCREEN_HEIGHT ) return;// only sky
			if( y1 < 0 && y2 < 0 ) {// only ground
				g.fillRect( 0, 0, GameView.SCREEN_WIDTH, GameView.SCREEN_HEIGHT );
				return;
			}
		}
		// clip line
		if( x1 < 0 ) {
			y1 -= x1 * (y2 - y1) / (x2 - x1);
			x1 = 0;
		}
		if( x2 >= GameView.SCREEN_WIDTH ) {
			y2 -= (x2 - GameView.SCREEN_WIDTH) * (y1 - y2) / (x1 - x2);
			x2 = GameView.SCREEN_WIDTH;
		}
		if( y1 == y2 && x1 == x2 ) return;// visible only 1 corner point
		if( y1 < 0 ) {
			x1 -= y1 * (x2 - x1) / (y2 - y1);
			y1 = 0;
		}
		if( y2 > GameView.SCREEN_HEIGHT ) {
			x2 -= ((y2 - GameView.SCREEN_HEIGHT) * (x1 - x2)) / (y1 - y2);
			y2 = GameView.SCREEN_HEIGHT;
		}
		// draw polygon
		// draw sloping ground section, 4 points maximum
		if( is_upside_down ) {
			g.fillRect( 0, 0, GameView.SCREEN_WIDTH, Math.min( y1, y2 ) );
		} else {
			final int y = Math.max( y1, y2 );
			g.fillRect( 0, y, GameView.SCREEN_WIDTH, GameView.SCREEN_HEIGHT - y );
		}
		if( y1 != y2 ) {
			final int[] xp = mXPoints;// for speed optimization
			final int[] yp = mYPoints;// for speed optimization
			int sides = 0;// 5 sides maximum
			//if( (isUpsideDown && (y1 > y2)) || (!isUpsideDown && (y1 < y2)) )
			if( is_upside_down ^ (y1 < y2) ) {// ground on left area of the screen
				if( x1 > 0 ) {
					xp[sides] = 0;
					yp[sides++] = y1;
				}
				xp[sides] = x1;
				yp[sides++] = y1;
				xp[sides] = x2;
				yp[sides++] = y2;
				if( x2 > 0 ) {
					xp[sides] = 0;
					yp[sides++] = y2;
				}
			} else {// ground on right area of the screen
				if( x1 < GameView.SCREEN_WIDTH ) {
					xp[sides] = GameView.SCREEN_WIDTH;
					yp[sides++] = y1;
				}
				xp[sides] = x1;
				yp[sides++] = y1;
				xp[sides] = x2;
				yp[sides++] = y2;
				if (x2 < GameView.SCREEN_WIDTH) {
					xp[sides] = GameView.SCREEN_WIDTH;
					yp[sides++] = y2;
				}
			}
			//if( sides > 2 )
				Support.fillPolygon( g, xp, yp, sides );
		}
		// horizon's line
		//g.setColor( Color.GRAY );
		//g.drawLine( x1, z1, x2, z2 );
	}
	/** Uses view matrix */
	public static void drawEnvironment(final Graphics g, final Camera camera, final Model[] scenery) {
		//final int max_point_count = 7;
		final int screen_coordinates_x[] = mScreenCoordinatesX;
		final int screen_coordinates_y[] = mScreenCoordinatesY;
		final int mr = 3;
		final int camera_z = camera.mZ;
		final int[] m = camera.mMatrix;
		// distance to objects do not changes, they are very far
		// so scenery x and y coordinates relative to the view point
		for( int s = 0; s < scenery.length; s++ ) {
			final Model so = scenery[s];
			boolean is_visible = true;
			for( int i = 0; i < so.mCoords.length; i++ ) {
				final Coord3D coord = so.mCoords[i];
				final int rz = coord.mZ - camera_z;
				final int y = (coord.mX * m[1 * mr + 0] + coord.mY * m[1 * mr + 1] + rz * m[1 * mr + 2]);
				if( y < 0 ) {// infront of the screen
					is_visible = false;
					break;
				}
				final int x = (coord.mX * m[0 * mr + 0] + coord.mY * m[0 * mr + 1] + rz * m[0 * mr + 2]);
				final int z = (coord.mX * m[2 * mr + 0] + coord.mY * m[2 * mr + 1] + rz * m[2 * mr + 2]);
				// to screen coordinates
				final int scale = (y + (CAMERA_FOV << Camera.PRECISION_LOG2)) / SCALE;
				screen_coordinates_x[i] = GameView.SCREEN_HALFWIDTH + (x / scale);
				screen_coordinates_y[i] = LOOK_Y - (z / scale);
			}
			if( is_visible ) {// draw scenery object
				//final int max_polygon_point_count = 4;
				final int x_points[] = mXPoints;
				final int y_points[] = mYPoints;
				for( int i = 0; i < so.mPolygons.length; i ++ ) {
					final Polygon p = so.mPolygons[i];
					for( int n = 0; n < p.mCoordIndexes.length; n++ ) {
						x_points[n] = screen_coordinates_x[p.mCoordIndexes[n]];
						y_points[n] = screen_coordinates_y[p.mCoordIndexes[n]];
					}
					g.setColor( p.mColor );
					Support.fillPolygon( g, x_points, y_points, p.mCoordIndexes.length );
				}
			}
		}
	}
	private static void drawWheel(final Graphics g, final OpponentCar car, int sx, int sy, int sz) {
		//final int point_count = car.mWheelY.length;
		final int[] x_points = mXPoints;// left side and right side
		final int[] y_points = mYPoints;// left side and right side
		final int wx_left = sx - Car.WHEEL_WIDTH / 2;
		final int wx_right = sx + Car.WHEEL_WIDTH / 2;
		final int[] m = mTransformMatrix;// for speed optimization
		final int wx_left_mult_res_0 = wx_left * m[0] + m[3];
		final int wx_left_mult_res_4 = wx_left * m[4] + m[7];
		final int wx_left_mult_res_8 = wx_left * m[8] + m[11];
		final int wx_right_mult_res_0 = wx_right * m[0] + m[3];
		final int wx_right_mult_res_4 = wx_right * m[4] + m[7];
		final int wx_right_mult_res_8 = wx_right * m[8] + m[11];
		//boolean isVisible = true;
		for( int i = 0; i < OpponentCar.WHEEL_CIRCLE_POINTS; i++ ) {
			final int wy = sy + car.mWheelY[i];
			final int wz = car.mWheelZ[i] + sz;
			// rotate and shift coordinates
			final int xt = wy * m[1] + wz * m[2];
			final int yt = wy * m[5] + wz * m[6];
			final int zt = wy * m[9] + wz * m[10];
			final int x_left = wx_left_mult_res_0 + xt;
			final int y_left = wx_left_mult_res_4 + yt;
			final int z_left = wx_left_mult_res_8 + zt;
			final int x_right = wx_right_mult_res_0 + xt;
			final int y_right = wx_right_mult_res_4 + yt;
			final int z_right = wx_right_mult_res_8 + zt;
			if( y_left < 0 || y_right < 0 )// infront of the screen
				return;
			// to screen coordinates
			final int scale_left = (y_left + (CAMERA_FOV << Camera.PRECISION_LOG2)) / SCALE;
			x_points[i] = GameView.SCREEN_HALFWIDTH + (x_left / scale_left);
			y_points[i] = LOOK_Y - (z_left / scale_left);
			final int scale_right = (y_right + (CAMERA_FOV << Camera.PRECISION_LOG2)) / SCALE;
			x_points[i + OpponentCar.WHEEL_CIRCLE_POINTS] = GameView.SCREEN_HALFWIDTH + (x_right / scale_right);
			y_points[i + OpponentCar.WHEEL_CIRCLE_POINTS] = LOOK_Y - (z_right / scale_right);
		}
		/*if( isVisible ) */{
			// tyre
			g.setColor( car.COLOR_TYRE );
			final int[] xp = mXPoints4;
			final int[] yp = mYPoints4;
			for( int i = 0; i < OpponentCar.WHEEL_CIRCLE_POINTS - 1; i++ ) {
				xp[0] = x_points[i];
				yp[0] = y_points[i];
				xp[1] = x_points[i + 1];
				yp[1] = y_points[i + 1];
				xp[2] = x_points[i + 1 + OpponentCar.WHEEL_CIRCLE_POINTS];
				yp[2] = y_points[i + 1 + OpponentCar.WHEEL_CIRCLE_POINTS];
				xp[3] = x_points[i + OpponentCar.WHEEL_CIRCLE_POINTS];
				yp[3] = y_points[i + OpponentCar.WHEEL_CIRCLE_POINTS];
				Support.fillPolygon( g, xp, yp, 4 );
				//g.drawPolygon( xp, yp, 4 );
			}
			xp[0] = x_points[OpponentCar.WHEEL_CIRCLE_POINTS - 1];
			yp[0] = y_points[OpponentCar.WHEEL_CIRCLE_POINTS - 1];
			xp[1] = x_points[0];
			yp[1] = y_points[0];
			xp[2] = x_points[OpponentCar.WHEEL_CIRCLE_POINTS];
			yp[2] = y_points[OpponentCar.WHEEL_CIRCLE_POINTS];
			xp[3] = x_points[OpponentCar.WHEEL_CIRCLE_POINTS - 1 + OpponentCar.WHEEL_CIRCLE_POINTS];
			yp[3] = y_points[OpponentCar.WHEEL_CIRCLE_POINTS - 1 + OpponentCar.WHEEL_CIRCLE_POINTS];
			Support.fillPolygon( g, xp, yp, 4 );
			//g.drawPolygon( xp, yp, 4 );
			g.setColor( car.COLOR_WHEEL );
			int i13 = OpponentCar.WHEEL_CIRCLE_POINTS >> 2;
			int i23 = OpponentCar.WHEEL_CIRCLE_POINTS - i13;
			if( (x_points[0] - x_points[i23]) * (y_points[i13] - y_points[i23]) -
				(y_points[0] - y_points[i23]) * (x_points[i13] - x_points[i23]) > 0 ) {
				Support.fillPolygon( g, x_points, y_points, OpponentCar.WHEEL_CIRCLE_POINTS );
				//g.drawPolygon( xPoints, yPoints, OpponentCar.WHEEL_CIRCLE_POINTS );
			}
			i13 += OpponentCar.WHEEL_CIRCLE_POINTS;
			i23 += OpponentCar.WHEEL_CIRCLE_POINTS;
			if( (x_points[OpponentCar.WHEEL_CIRCLE_POINTS] - x_points[i13]) * (y_points[i23] - y_points[i13]) -
				(y_points[OpponentCar.WHEEL_CIRCLE_POINTS] - y_points[i13]) * (x_points[i23] - x_points[i13]) > 0 ) {
				for( int i = 0; i < OpponentCar.WHEEL_CIRCLE_POINTS; i++ ) {
					x_points[i] = x_points[i + OpponentCar.WHEEL_CIRCLE_POINTS];
					y_points[i] = y_points[i + OpponentCar.WHEEL_CIRCLE_POINTS];
				}
				Support.fillPolygon( g, x_points, y_points, OpponentCar.WHEEL_CIRCLE_POINTS );
				//g.drawPolygon( xPoints, yPoints, OpponentCar.WHEEL_CIRCLE_POINTS );
			}
			// calculate disk polygon coordinates
			for( int i = 0; i < (OpponentCar.WHEEL_CIRCLE_POINTS >> 1); i++ ) {
				int r = i + (OpponentCar.WHEEL_CIRCLE_POINTS >> 1);
				if( r >= OpponentCar.WHEEL_CIRCLE_POINTS ) r -= OpponentCar.WHEEL_CIRCLE_POINTS;
				final int xshift = (x_points[r] - x_points[i]) >> 2;
				x_points[i] += xshift;
				x_points[r] -= xshift;
				final int yshift = (y_points[r] - y_points[i]) >> 2;
				y_points[i] += yshift;
				y_points[r] -= yshift;
			}
			// disk
			g.setColor( car.COLOR_DISK );
			Support.fillPolygon( g, x_points, y_points, OpponentCar.WHEEL_CIRCLE_POINTS );
			//g.drawPolygon( xPoints, yPoints, OpponentCar.WHEEL_CIRCLE_POINTS );
		}
	}
	/**
	 * @param mCamera - camera matrix and coordinates
	 * @param car - object matrix and object's CG coordinates
	 * @return mTransformMatrix to transform object's coordinates to camera coordinates
	 */
	private static void calculateTransformMatrix(final Camera camera, final Car car) {
		// to speed
		final int[] vm = camera.mMatrix;
		final int[] om = car.mMatrix;
		final int dx = (car.mCG_X) - camera.mX;
		final int dy = (car.mCG_Y) - camera.mY;
		final int dz = (car.mCG_Z) - camera.mZ;
		final int[] tm = mTransformMatrix;
		final int tmr = 4;
		final int vmr = 3;
		//
		tm[0 * tmr + 0] = (om[0 * vmr + 0] * vm[0 * vmr + 0] + om[0 * vmr + 1] * vm[0 * vmr + 1] + om[0 * vmr + 2] * vm[0 * vmr + 2]) / Camera.PRECISION;
		tm[0 * tmr + 1] = (om[1 * vmr + 0] * vm[0 * vmr + 0] + om[1 * vmr + 1] * vm[0 * vmr + 1] + om[1 * vmr + 2] * vm[0 * vmr + 2]) / Camera.PRECISION;
		tm[0 * tmr + 2] = (om[2 * vmr + 0] * vm[0 * vmr + 0] + om[2 * vmr + 1] * vm[0 * vmr + 1] + om[2 * vmr + 2] * vm[0 * vmr + 2]) / Camera.PRECISION;
		tm[0 * tmr + 3] = (             dx * vm[0 * vmr + 0] +              dy * vm[0 * vmr + 1] +              dz * vm[0 * vmr + 2]);
		//
		tm[1 * tmr + 0] = (om[0 * vmr + 0] * vm[1 * vmr + 0] + om[0 * vmr + 1] * vm[1 * vmr + 1] + om[0 * vmr + 2] * vm[1 * vmr + 2]) / Camera.PRECISION;
		tm[1 * tmr + 1] = (om[1 * vmr + 0] * vm[1 * vmr + 0] + om[1 * vmr + 1] * vm[1 * vmr + 1] + om[1 * vmr + 2] * vm[1 * vmr + 2]) / Camera.PRECISION;
		tm[1 * tmr + 2] = (om[2 * vmr + 0] * vm[1 * vmr + 0] + om[2 * vmr + 1] * vm[1 * vmr + 1] + om[2 * vmr + 2] * vm[1 * vmr + 2]) / Camera.PRECISION;
		tm[1 * tmr + 3] = (             dx * vm[1 * vmr + 0] +              dy * vm[1 * vmr + 1] +              dz * vm[1 * vmr + 2]);
		//
		tm[2 * tmr + 0] = (om[0 * vmr + 0] * vm[2 * vmr + 0] + om[0 * vmr + 1] * vm[2 * vmr + 1] + om[0 * vmr + 2] * vm[2 * vmr + 2]) / Camera.PRECISION;
		tm[2 * tmr + 1] = (om[1 * vmr + 0] * vm[2 * vmr + 0] + om[1 * vmr + 1] * vm[2 * vmr + 1] + om[1 * vmr + 2] * vm[2 * vmr + 2]) / Camera.PRECISION;
		tm[2 * tmr + 2] = (om[2 * vmr + 0] * vm[2 * vmr + 0] + om[2 * vmr + 1] * vm[2 * vmr + 1] + om[2 * vmr + 2] * vm[2 * vmr + 2]) / Camera.PRECISION;
		tm[2 * tmr + 3] = (             dx * vm[2 * vmr + 0] +              dy * vm[2 * vmr + 1] +              dz * vm[2 * vmr + 2]);
	}
	/**
	 * before calling this function must be calculated:<br>
	 * - camera x,y,z and angles
	 * - opponent car x,y,z and angles
	 * - mViewMatrix
	 */
	private static void drawOpponentCar(final Graphics g, final Camera camera, final OpponentCar car, boolean isDrawShadow) {
		// calculate transform matrix to render opponent car
		calculateTransformMatrix( camera, car );
		// for speed optimization
		final Coord3D[] coords = car.mHull.mCoords;
		//final int max_point_count = 18;// or coords.length
		// coords.length + 4 extra points to store shadow coordinates
		final int screen_coordinates_x[] = mScreenCoordinatesX;
		final int screen_coordinates_y[] = mScreenCoordinatesY;
		final int tmr = 4;
		final int[] tm = mTransformMatrix;
		final int mr = 3;
		final int[] m = car.mMatrix;
		// first 4 points use to calculate the car shadow position
		for( int i = 0; i < 4; i++ ) {
			final Coord3D c = coords[i];
			final int y_fp = c.mX * tm[1 * tmr + 0] + c.mY * tm[1 * tmr + 1] + c.mZ * tm[1 * tmr + 2] + tm[1 * tmr + 3];
			if( y_fp < 0 )// infront of the screen
				return;
			final int x_fp = c.mX * tm[0 * tmr + 0] + c.mY * tm[0 * tmr + 1] + c.mZ * tm[0 * tmr + 2] + tm[0 * tmr + 3];
			final int z_fp = c.mX * tm[2 * tmr + 0] + c.mY * tm[2 * tmr + 1] + c.mZ * tm[2 * tmr + 2] + tm[2 * tmr + 3];
			// to screen coordinates
			final int scale = (y_fp + (CAMERA_FOV << Camera.PRECISION_LOG2)) / SCALE;
			screen_coordinates_x[i] = GameView.SCREEN_HALFWIDTH + (x_fp / scale);
			screen_coordinates_y[i] = LOOK_Y - (z_fp / scale);
			// calculate shadow screen coordinates
			// local to world about the car CG
			final int shadow_world_x = (car.mCG_X) + ((c.mX * m[0 * mr + 0] + c.mY * m[1 * mr + 0] + c.mZ * m[2 * mr + 0]) >> Camera.PRECISION_LOG2);
			final int shadow_world_y = (car.mCG_Y) + ((c.mX * m[0 * mr + 1] + c.mY * m[1 * mr + 1] + c.mZ * m[2 * mr + 1]) >> Camera.PRECISION_LOG2);
			final int shadow_world_z = ((car.interpolateHeight( sTrack, shadow_world_x, shadow_world_y, false )) >> Camera.PRECISION_LOG2);
			//
			final int shadow_local_x = shadow_world_x - camera.mX;
			final int shadow_local_y = shadow_world_y - camera.mY;
			final int shadow_local_z = shadow_world_z - camera.mZ;
			//
			final int[] cm = camera.mMatrix;
			final int shadow_y_fp = shadow_local_x * cm[1 * mr + 0] + shadow_local_y * cm[1 * mr + 1] + shadow_local_z * cm[1 * mr + 2];
			if( shadow_y_fp < 0 )// infront of the screen
				return;
			final int shadow_x_fp = shadow_local_x * cm[0 * mr + 0] + shadow_local_y * cm[0 * mr + 1] + shadow_local_z * cm[0 * mr + 2];
			final int shadow_z_fp = shadow_local_x * cm[2 * mr + 0] + shadow_local_y * cm[2 * mr + 1] + shadow_local_z * cm[2 * mr + 2];
			// to screen coordinates
			final int shadow_scale = (shadow_y_fp + (CAMERA_FOV << Camera.PRECISION_LOG2)) / SCALE;
			screen_coordinates_x[coords.length + i] = GameView.SCREEN_HALFWIDTH + (shadow_x_fp / shadow_scale);
			screen_coordinates_y[coords.length + i] = LOOK_Y - (shadow_z_fp / shadow_scale);
		}
		for( int i = 4; i < coords.length; i++ ) {
			final Coord3D c = coords[i];
			final int y = c.mX * tm[1 * tmr + 0] + c.mY * tm[1 * tmr + 1] + c.mZ * tm[1 * tmr + 2] + tm[1 * tmr + 3];
			if( y < 0 )// infront of the screen
				return;
			final int x = c.mX * tm[0 * tmr + 0] + c.mY * tm[0 * tmr + 1] + c.mZ * tm[0 * tmr + 2] + tm[0 * tmr + 3];
			final int z = c.mX * tm[2 * tmr + 0] + c.mY * tm[2 * tmr + 1] + c.mZ * tm[2 * tmr + 2] + tm[2 * tmr + 3];
			// to screen coordinates
			final int scale = (y + (CAMERA_FOV << Camera.PRECISION_LOG2)) / SCALE;
			screen_coordinates_x[i] = GameView.SCREEN_HALFWIDTH + (x / scale);
			screen_coordinates_y[i] = LOOK_Y - (z / scale);
		}
		// draw polygons
		//final int max_polygon_point_count = 8;
		final int x_points[] = mXPoints;
		final int y_points[] = mYPoints;
		// render shadow
		x_points[0] = screen_coordinates_x[coords.length + 0];
		y_points[0] = screen_coordinates_y[coords.length + 0];
		x_points[1] = screen_coordinates_x[coords.length + 3];
		y_points[1] = screen_coordinates_y[coords.length + 3];
		x_points[2] = screen_coordinates_x[coords.length + 2];
		y_points[2] = screen_coordinates_y[coords.length + 2];
		x_points[3] = screen_coordinates_x[coords.length + 1];
		y_points[3] = screen_coordinates_y[coords.length + 1];
		// check visibility
		if( isDrawShadow &&
			(x_points[0] - x_points[1]) * (y_points[2] - y_points[1]) -
			(y_points[0] - y_points[1]) * (x_points[2] - x_points[1]) > 0 )
		{
			g.setColor( 0x202020 );
			Support.fillPolygon( g, x_points, y_points, 4 );
		}
		// render car
		boolean is_left_side_visible =
			(screen_coordinates_x[car.left_side_0] - screen_coordinates_x[car.left_side_1]) *
			(screen_coordinates_y[car.left_side_2] - screen_coordinates_y[car.left_side_1]) -
			(screen_coordinates_y[car.left_side_0] - screen_coordinates_y[car.left_side_1]) *
			(screen_coordinates_x[car.left_side_2] - screen_coordinates_x[car.left_side_1]) > 0;
		boolean is_right_side_visible =
			(screen_coordinates_x[car.right_side_0] - screen_coordinates_x[car.right_side_1]) *
			(screen_coordinates_y[car.right_side_2] - screen_coordinates_y[car.right_side_1]) -
			(screen_coordinates_y[car.right_side_0] - screen_coordinates_y[car.right_side_1]) *
			(screen_coordinates_x[car.right_side_2] - screen_coordinates_x[car.right_side_1]) > 0;

		if( is_left_side_visible ) {
			drawWheel( g, car, Car.WHEEL_RIGHT_X, Car.WHEEL_FRONT_Y, car.mFrontRightWheelBottom );
			drawWheel( g, car, Car.WHEEL_RIGHT_X, Car.WHEEL_REAR_Y, car.mRearRightWheelBottom );
		} else if( is_right_side_visible ) {
			drawWheel( g, car, Car.WHEEL_LEFT_X, Car.WHEEL_FRONT_Y, car.mFrontLeftWheelBottom );
			drawWheel( g, car, Car.WHEEL_LEFT_X, Car.WHEEL_REAR_Y, car.mRearLeftWheelBottom );
		} else {
			drawWheel( g, car, Car.WHEEL_RIGHT_X, Car.WHEEL_FRONT_Y, car.mFrontRightWheelBottom );
			drawWheel( g, car, Car.WHEEL_RIGHT_X, Car.WHEEL_REAR_Y, car.mRearRightWheelBottom );
			drawWheel( g, car, Car.WHEEL_LEFT_X, Car.WHEEL_FRONT_Y, car.mFrontLeftWheelBottom );
			drawWheel( g, car, Car.WHEEL_LEFT_X, Car.WHEEL_REAR_Y, car.mRearLeftWheelBottom );
		}
		//
		final Polygon[] polygons = car.mHull.mPolygons;
		for( int i = 0; i < polygons.length; i ++ ) {
			final Polygon p = polygons[i];
			int index = p.mCoordIndexes[0];
			x_points[0] = screen_coordinates_x[index];
			y_points[0] = screen_coordinates_y[index];
			index = p.mCoordIndexes[1];
			x_points[1] = screen_coordinates_x[index];
			y_points[1] = screen_coordinates_y[index];
			index = p.mCoordIndexes[2];
			x_points[2] = screen_coordinates_x[index];
			y_points[2] = screen_coordinates_y[index];
			// check visibility
			if( (x_points[0] - x_points[1]) * (y_points[2] - y_points[1]) -
				(y_points[0] - y_points[1]) * (x_points[2] - x_points[1]) > 0 ) {
				for( int n = 3; n < p.mCoordIndexes.length; n++ ) {
					index = p.mCoordIndexes[n];
					x_points[n] = screen_coordinates_x[index];
					y_points[n] = screen_coordinates_y[index];
				}
				g.setColor( p.mColor );
				Support.fillPolygon( g, x_points, y_points, p.mCoordIndexes.length );
				g.setColor( 0x202020 );
				Support.drawPolygon( g, x_points, y_points, p.mCoordIndexes.length );
			}
		}
		//
		if( is_left_side_visible ) {
			drawWheel( g, car, Car.WHEEL_LEFT_X, Car.WHEEL_FRONT_Y, car.mFrontLeftWheelBottom );
			drawWheel( g, car, Car.WHEEL_LEFT_X, Car.WHEEL_REAR_Y, car.mRearLeftWheelBottom );
		} else if( is_right_side_visible ) {
			drawWheel( g, car, Car.WHEEL_RIGHT_X, Car.WHEEL_FRONT_Y, car.mFrontRightWheelBottom );
			drawWheel( g, car, Car.WHEEL_RIGHT_X, Car.WHEEL_REAR_Y, car.mRearRightWheelBottom );
		}
	}
	//
	public static void setTrack(final Track track) {
		sTrack = track;
		// calculating rough section count
		final TrackPiece[] tp = track.mPieces;
		int rough_count = tp.length;
		for( int piece = 0; piece < tp.length; piece++ ) {
			Coord3D[] coords = tp[piece].mCoords;
			final int length = coords.length - 2;
			for( int c = 0; c < length; ) {
				int next = c + 2;
				if( Math.abs( coords[c].mZ - coords[next].mZ ) >= STEEP_Z_DIFFERENCE) {
					rough_count++;
					if( next < length )
						rough_count++;
				}
				c = next;
			}
		}
		//
		sSections = new Section[rough_count + TRACK_ACCURATE_VIEW];
		sAuxSections = new Section[sSections.length];
		//
		Section section = null;
		int count = 0;
		int coord_index = sTrack.mNumTrackCoords;
		for( int piece = tp.length; piece-- > 0; ) {
			Coord3D[] coords = tp[piece].mCoords;
			// search sections with steep angle
			final int length = coords.length - 2;
			for( int c = length; c > 0; ) {
				coord_index -= 2;
				int prev = c - 2;
				if( Math.abs( coords[c].mZ - coords[prev].mZ ) >= STEEP_Z_DIFFERENCE) {
					if( c < length ) {
						section = new Section( coord_index + 2, coords[c], coords[c + 1], tp[piece], false, section, false );
						sSections[count++] = section;
					}
					section = new Section( coord_index, coords[prev], coords[prev + 1], tp[piece], true, section, false );
					sSections[count++] = section;
				}
				c = prev;
			}
			section = new Section( coord_index, coords[0], coords[1], tp[piece], false, section, false );
			sSections[count++] = section;
		}
		sSections[0].mNextSection = sSections[count - 1];
		for( int i = 0; i < TRACK_ACCURATE_VIEW; i++ ) {
			sSections[count++] = new Section( 0, null, null, null, false, null, true );
		}
	}
	public static void freeTrack() {
		sTrack = null;
		sSections = null;
		sAuxSections = null;
	}
	/** Clip near coordinates by y = 0 */
	private static void clip(final Coord3D far, final Coord3D near) {
		final int dy = (near.mY - far.mY) >> Camera.PRECISION_LOG2;
		final int y = near.mY >> Camera.PRECISION_LOG2;
		near.mX += (((far.mX - near.mX) >> Camera.PRECISION_LOG2) * y / dy) << Camera.PRECISION_LOG2;
		near.mZ += (((far.mZ - near.mZ) >> Camera.PRECISION_LOG2) * y / dy) << Camera.PRECISION_LOG2;
		near.mY = 0;
	}
	/** @return true, if far section behind the camera */
	private static boolean drawSection(final Graphics g, Section s, Section n, final Camera camera, final OpponentCar opponentCar) {
		final boolean is_near = (s.mIsNear & n.mIsNear);// | sIsPreview;// if preview, uses sections with mIsNear flag set
		final boolean is_reverse = s.mViewDistance_d64 > n.mViewDistance_d64;
		if( is_reverse ) {
			Section t = s;
			s = n;
			n = t;
		}
		//
		if( is_near ) {// clip by plane y = 0
			if( s.mLeftTop.mY < 0 && n.mLeftTop.mY >= 0 ) clip( n.mLeftTop, s.mLeftTop );
			if( s.mRightTop.mY < 0 && n.mRightTop.mY >= 0 ) clip( n.mRightTop, s.mRightTop );
			if( s.mLeftBottom.mY < 0 && n.mLeftBottom.mY >= 0 ) clip( n.mLeftBottom, s.mLeftBottom );
			if( s.mRightBottom.mY < 0 && n.mRightBottom.mY >= 0 ) clip( n.mRightBottom, s.mRightBottom );
		}
		//
		if( s.mLeftTop.mY >= 0 && s.mRightTop.mY >= 0 && n.mLeftTop.mY >= 0 && n.mRightTop.mY >= 0 )
		{// clip by plane y = 0
			if( n.mLeftBottom.mY < 0 ) clip( n.mLeftTop, n.mLeftBottom );
			if( n.mRightBottom.mY < 0 ) clip( n.mRightTop, n.mRightBottom );
			if( s.mLeftBottom.mY < 0 ) clip( s.mLeftTop, s.mLeftBottom );
			if( s.mRightBottom.mY < 0 ) clip( s.mRightTop, s.mRightBottom );
			//
			final int[] xp = mXPoints;// for speed optimization
			final int[] yp = mYPoints;// for speed optimization
			s.worldToScreen();
			n.worldToScreen();
			// draw top section
			int color = is_reverse ? n.mPiece.mRoadColor : s.mPiece.mRoadColor;
			final boolean is_steep = is_reverse ? n.mIsSteep : s.mIsSteep;
			if( is_steep )// rapid ascend or descend
				color = 0;
			g.setColor( color );
			// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
			// orientation > 0 : left
			// orientation < 0 : right
			// orientation == 0 : on
			// use reverse orientation since display y axis is oriented downwards
			/*int is_visible =// there is a problem with a skew surface, i.e. piece[12] of the Little Ramp
					(n.mRightTopX - s.mLeftTopX) * (s.mRightTopY - s.mLeftTopY) -
					(n.mRightTopY - s.mLeftTopY) * (s.mRightTopX - s.mLeftTopX);
			if( is_reverse ) is_visible = -is_visible;
			if( is_visible < 0 ) {
				is_visible =//
					(n.mLeftTopY - s.mRightTopY) * (s.mLeftTopX - s.mRightTopX) -
					(n.mLeftTopX - s.mRightTopX) * (s.mLeftTopY - s.mRightTopY);
				if( is_reverse ) is_visible = -is_visible;
			}
			if( is_visible > 0 ) */{// draw top section. always overlap by sides, so visibility check do not needed
				xp[0] = s.mLeftTopX;
				yp[0] = s.mLeftTopY;
				xp[1] = n.mLeftTopX;
				yp[1] = n.mLeftTopY;
				xp[2] = n.mRightTopX;
				yp[2] = n.mRightTopY;
				xp[3] = s.mRightTopX;
				yp[3] = s.mRightTopY;
				Support.fillPolygon( g, xp, yp, 4 );
				if( is_near ) {
					color = sTrack.mRoadLineColorDark;
					if( ((s.mCoordIndex & 2) != 0) ^ is_reverse ) {
						if( ! is_steep ) {// draw marking line
							g.setColor( 0xc0c0c0 );
							g.drawLine( (s.mRightTopX + s.mLeftTopX) >> 1,
									(s.mRightTopY + s.mLeftTopY) >> 1,
									(n.mRightTopX + n.mLeftTopX) >> 1,
									(n.mRightTopY + n.mLeftTopY) >> 1 );
						}
						color = sTrack.mRoadLineColorLight;
					}
					// draw side lines
					g.setColor( color );
					g.drawLine( s.mLeftTopX, s.mLeftTopY, n.mLeftTopX, n.mLeftTopY );
					g.drawLine( s.mRightTopX, s.mRightTopY, n.mRightTopX, n.mRightTopY );
				}
				if( n.mCoordIndex == 0 ) {// draw start line
					g.setColor( sTrack.mRoadStartLineColor );
					g.drawLine( n.mLeftTopX, n.mLeftTopY, n.mRightTopX, n.mRightTopY );
				}
			}// if draw top section
			//
			if( ! opponentCar.mIsRendered ) {
				if( n.mCoordIndex == opponentCar.mCoordIndex ) {
					drawOpponentCar( g, camera, opponentCar, ! is_steep );
					opponentCar.mIsRendered = true;
				}
			}
			// draw left section
			// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
			// orientation > 0 : left
			// orientation < 0 : right
			// orientation == 0 : on
			// use reverse orientation since display y axis is oriented downwards
			int is_visible =//
				(n.mLeftTopX - s.mLeftBottomX) * (s.mLeftTopY - s.mLeftBottomY) -
				(n.mLeftTopY - s.mLeftBottomY) * (s.mLeftTopX - s.mLeftBottomX);
			if( is_visible == 0 ) {
				is_visible =//
					(s.mLeftBottomX - n.mLeftTopX) * (n.mLeftBottomY - n.mLeftTopY) -
					(s.mLeftBottomY - n.mLeftTopY) * (n.mLeftBottomX - n.mLeftTopX);
			}
			if( is_reverse ) is_visible = -is_visible;
			if( is_visible > 0 ) {
				xp[0] = s.mLeftBottomX;
				yp[0] = s.mLeftBottomY;
				xp[1] = n.mLeftBottomX;
				yp[1] = n.mLeftBottomY;
				xp[2] = n.mLeftTopX;
				yp[2] = n.mLeftTopY;
				xp[3] = s.mLeftTopX;
				yp[3] = s.mLeftTopY;
				color = is_reverse ? n.mPiece.mSideColor : s.mPiece.mSideColor;
				g.setColor( color );
				Support.fillPolygon( g, xp, yp, 4 );
				//g.setColor( sTrack.mRoadLineColorDark );
				if( is_near ) {// draw vertical line
					g.setColor( sTrack.mRoadLineColorDark );
					if( ! sIsPreview ) g.drawLine( n.mLeftBottomX, n.mLeftBottomY, n.mLeftTopX, n.mLeftTopY );// far
				}// else {// draw side line
				//	g.drawLine( s.mLeftTopX, s.mLeftTopY, n.mLeftTopX, n.mLeftTopY );
				//}
			}// if draw left section
			// draw right section
			// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
			// orientation > 0 : left
			// orientation < 0 : right
			// orientation == 0 : on
			// use reverse orientation since display y axis is oriented downwards
			is_visible =//
				(n.mRightTopX - s.mRightBottomX) * (s.mRightTopY - s.mRightBottomY) -
				(n.mRightTopY - s.mRightBottomY) * (s.mRightTopX - s.mRightBottomX);
			if( is_visible == 0 ) {
				is_visible =//
					(s.mRightBottomX - n.mRightTopX) * (n.mRightBottomY - n.mRightTopY) -
					(s.mRightBottomY - n.mRightTopY) * (n.mRightBottomX - n.mRightTopX);
			}
			if( is_reverse ) is_visible = -is_visible;
			if( is_visible < 0 ) {
				xp[0] = s.mRightTopX;
				yp[0] = s.mRightTopY;
				xp[1] = n.mRightTopX;
				yp[1] = n.mRightTopY;
				xp[2] = n.mRightBottomX;
				yp[2] = n.mRightBottomY;
				xp[3] = s.mRightBottomX;
				yp[3] = s.mRightBottomY;
				color = is_reverse ? n.mPiece.mSideColor : s.mPiece.mSideColor;
				g.setColor( color );
				Support.fillPolygon( g, xp, yp, 4 );
				//g.setColor( sTrack.mRoadLineColorDark );
				if( is_near ) {// draw vertical line
					g.setColor( sTrack.mRoadLineColorDark );
					if( ! sIsPreview ) g.drawLine( n.mRightTopX, n.mRightTopY, n.mRightBottomX, n.mRightBottomY );// far
				}// else {// draw side line
				//	g.drawLine( s.mRightTopX, s.mRightTopY, n.mRightTopX, n.mRightTopY );
				//}
			}// if draw right section
		}// if this section is visible
		return n.mViewDistance_d64 < 0;
	}
	/** Draw track and opponent car */
	public static void drawTrack(final Graphics g, final Camera camera, final Car car, final OpponentCar opponentCar) {
		// set render order
		Section.prepareToCalculate( camera );
		//
		int min_distance = sTrack.mNumTrackCoords >> 1;
		int max_distance = min_distance;
		Section last_section = sSections[0];
		Section prev_section = last_section;
		int near_last_count = -1;
		int near_first_count = 0;
		Section prev_near_section = null;
		int piece = car.mPieceIndex;
		int index = car.mPieceCoordIndex;
		final int first_coord_index = car.mCoordIndex;
		int coordIndex = first_coord_index;
		int i = 0;
		do {
			final Section s = sSections[i];
			if( s.mIsNear ) {
				if( near_last_count < 0 ) near_first_count = i;
				near_last_count = i;
				s.mPiece = sTrack.mPieces[piece];
				final Coord3D[] coords = s.mPiece.mCoords;
				s.mLeft = coords[index++];
				s.mRight = coords[index++];
				s.mCoordIndex = coordIndex;
				s.mIsSteep = Math.abs( s.mLeft.mZ - coords[index].mZ ) >= STEEP_Z_DIFFERENCE;
				if( prev_near_section != null ) prev_near_section.mNextSection = s;
				prev_near_section = s;
				coordIndex += 2;
				if( coordIndex >= sTrack.mNumTrackCoords ) coordIndex -= sTrack.mNumTrackCoords;
				if( index >= coords.length - 2 ) {
					index = 0;
					if( ++piece >= sTrack.mPieces.length )
						piece = 0;
				}
				s.calculateCoordinates( camera );
			} else {
				int d = s.mCoordIndex - first_coord_index;
				if( d < 0 ) d += sTrack.mNumTrackCoords;
				if( d < TRACK_ACCURATE_VIEW * 2 ) {
					s.mViewDistance_d64 = OFF_DISTANCE;
				} else {
					if( d < min_distance ) {
						min_distance = d;
						last_section = s;
					} else if( d > max_distance ) {
						max_distance = d;
						prev_section = s;
					}
					s.calculateCoordinates( camera );
				}
			}
		} while( ++i < sSections.length );
		sSections[near_last_count].mNextSection = last_section;
		// must be restored
		sSectionStore = prev_section;
		sSectionStoreNext = prev_section.mNextSection;
		prev_section.mNextSection = sSections[near_first_count];
		// sort sections
		System.arraycopy( sSections, 0, sAuxSections, 0, sAuxSections.length );
		mergeSort( sAuxSections, sSections, 0, sSections.length, 0 );
		//
		i = sSections.length;
		do {
			Section s = sSections[--i];
			if( s.mViewDistance_d64 == OFF_DISTANCE )
				break;
			Section n = s.mNextSection;
			if( drawSection( g, s, n, camera, opponentCar ) )
				break;//if( n.mViewDistance_d64 < 0 )// other sections will be behind the camera
		} while( i > 0 );
		if( ! opponentCar.mIsRendered ) {
			drawOpponentCar( g, camera, opponentCar, false );
		}
		// restore section order
		sSectionStore.mNextSection = sSectionStoreNext;
	}
	public static void drawTrackPreview(final Graphics g, final Camera camera, final OpponentCar opponentCar) {
		Section.prepareToCalculate( camera );
		//
		int i = sSections.length;
		do {
			final Section s = sSections[--i];
			if( s.mIsNear ) {
				s.mViewDistance_d64 = OFF_DISTANCE;
			} else {
				s.calculateCoordinates( camera );
			}
		} while( i > 0 );
		sSectionStore = sSections[0];
		sSectionStoreNext = sSectionStore.mNextSection;
		// sort sections
		System.arraycopy( sSections, 0, sAuxSections, 0, sAuxSections.length );
		mergeSort( sAuxSections, sSections, 0, sSections.length, 0 );
		//
		i = sSections.length;
		TrackPiece tp = null;
		Section s = sSections[0];
		Section n = sSections[1];
		do {
			final Section rs = sSections[--i];
			if( tp != rs.mPiece ) {// to skip steep sections
				tp = rs.mPiece;
				final Coord3D[] coords = tp.mCoords;
				if( rs.mViewDistance_d64 < rs.mNextSection.mViewDistance_d64 ) {
					n.mCoordIndex = rs.mCoordIndex + coords.length - 2;
					if( n.mCoordIndex >= sTrack.mNumTrackCoords ) n.mCoordIndex -= sTrack.mNumTrackCoords;
					n.mPiece = tp;
					n.mLeft = coords[coords.length - 2];
					n.mRight = coords[coords.length - 1];
					n.calculateCoordinates( camera );
					for( int k = coords.length - 4; k >= 0; k += -2 ) {
						s.mPiece = tp;
						s.mCoordIndex = rs.mCoordIndex + k;// uses only to choose color, so may be more than track's number of coordinates
						s.mLeft = coords[k];
						s.mRight = coords[k + 1];
						s.mIsSteep = Math.abs( s.mLeft.mZ - n.mLeft.mZ ) >= STEEP_Z_DIFFERENCE;
						//
						s.calculateCoordinates( camera );
						//
						drawSection( g, s, n, camera, opponentCar );
						//
						Section t = s;
						s = n;
						n = t;
					}
				} else {
					s.mCoordIndex = rs.mCoordIndex;
					s.mPiece = tp;
					s.mLeft = coords[0];
					s.mRight = coords[1];
					s.calculateCoordinates( camera );
					for( int k = 2; k < coords.length; ) {
						n.mCoordIndex = rs.mCoordIndex + k;// uses only to choose color, so may be more than track's number of coordinates
						n.mPiece = tp;
						n.mLeft = coords[k++];
						n.mRight = coords[k++];
						//
						s.mIsSteep = Math.abs( s.mLeft.mZ - n.mLeft.mZ ) >= STEEP_Z_DIFFERENCE;
						//
						n.calculateCoordinates( camera );
						//
						drawSection( g, s, n, camera, opponentCar );
						//
						Section t = s;
						s = n;
						n = t;
					}
				}
			}
		} while( i > TRACK_ACCURATE_VIEW );
	}
}