public class Section implements MyComparable {
	/** camera.mZ * m[0 * mr + 2] */
	private static int sDz0 = 0;
	/** camera.mZ * m[1 * mr + 2] */
	private static int sDz1 = 0;
	/** camera.mZ * m[2 * mr + 2] */
	private static int sDz2 = 0;
	//
	/** to check start line and is odd or is even */
	public int mCoordIndex;
	public Coord3D mLeft;
	public Coord3D mRight;
	/** to access to the piece colors */
	public TrackPiece mPiece;
	public boolean mIsSteep;
	public Section mNextSection;
	public final boolean mIsNear;
	public final Coord3D mLeftTop = new Coord3D();
	public final Coord3D mLeftBottom = new Coord3D();
	public final Coord3D mRightTop = new Coord3D();
	public final Coord3D mRightBottom = new Coord3D();
	public int mLeftTopX = 0;
	public int mLeftTopY = 0;
	public int mLeftBottomX = 0;
	public int mLeftBottomY = 0;
	public int mRightTopX = 0;
	public int mRightTopY = 0;
	public int mRightBottomX = 0;
	public int mRightBottomY = 0;
	/** square of view distance */
	public int mViewDistance_d64 = 0;
	private boolean mIsNeedCalculateScreenCoords = true;	
	//
	public Section(int coordIndex, Coord3D left, Coord3D right, TrackPiece piece, boolean isSteep, Section nextSection, boolean isNear) {
		mLeft = left;
		mRight = right;
		mIsSteep = isSteep;
		mNextSection = nextSection;
		mPiece = piece;
		mCoordIndex = coordIndex;
		mIsNear = isNear;
	}
	/** Calculate general values before calculate coordinates */
	public static void prepareToCalculate(final Camera camera) {
		final int mr = 3;
		final int[] m = camera.mMatrix;
		sDz0 = camera.mZ * m[0 * mr + 2];
		sDz1 = camera.mZ * m[1 * mr + 2];
		sDz2 = camera.mZ * m[2 * mr + 2];
	}
	/** Before use call the prepareToCalculate function */
	public void calculateCoordinates(final Camera camera) {
		int dx, dy, dz, temp;
		final int mr = 3;
		// to speed
		final int[] m = camera.mMatrix;
		//
		mIsNeedCalculateScreenCoords = true;
		//
		dx = mLeft.mX - camera.mX;
		dy = mLeft.mY - camera.mY;
		dz = mLeft.mZ - camera.mZ;
		//
		temp = dx * m[0 * mr + 0] + dy * m[0 * mr + 1];
		mLeftTop.mX = temp + dz * m[0 * mr + 2];
		mLeftBottom.mX = temp - sDz0;
		temp = dx * m[1 * mr + 0] + dy * m[1 * mr + 1];
		mLeftTop.mY = temp + dz * m[1 * mr + 2];
		mLeftBottom.mY = temp - sDz1;
		temp = dx * m[2 * mr + 0] + dy * m[2 * mr + 1];
		mLeftTop.mZ = temp + dz * m[2 * mr + 2];
		mLeftBottom.mZ = temp - sDz2;
		//
		dx = mRight.mX - camera.mX;
		dy = mRight.mY - camera.mY;
		dz = mRight.mZ - camera.mZ;
		//
		temp = dx * m[0 * mr + 0] + dy * m[0 * mr + 1];
		mRightTop.mX = temp + dz * m[0 * mr + 2];
		mRightBottom.mX = temp - sDz0;
		temp = dx * m[1 * mr + 0] + dy * m[1 * mr + 1];
		mRightTop.mY = temp + dz * m[1 * mr + 2];
		mRightBottom.mY = temp - sDz1;
		temp = dx * m[2 * mr + 0] + dy * m[2 * mr + 1];
		mRightTop.mZ = temp + dz * m[2 * mr + 2];
		mRightBottom.mZ = temp - sDz2;
		//
		dx = mLeftTop.mX >> (Camera.PRECISION_LOG2 + 3);// +3 to prevent overflow
		dy = mLeftTop.mY >> (Camera.PRECISION_LOG2 + 3);
		temp = dx * dx + dy * dy;
		if( dy < 0 ) temp = -temp;
		mViewDistance_d64 = temp;
	}
	public void worldToScreen() {
		if( mIsNeedCalculateScreenCoords ) {
			mIsNeedCalculateScreenCoords = false;
			int scale = (mLeftTop.mY + (Render.CAMERA_FOV << Camera.PRECISION_LOG2)) / Render.SCALE;
			mLeftTopX = GameView.SCREEN_HALFWIDTH + (mLeftTop.mX / scale);
			mLeftTopY = Render.LOOK_Y - (mLeftTop.mZ / scale);
	
			scale = (mLeftBottom.mY + (Render.CAMERA_FOV << Camera.PRECISION_LOG2)) / Render.SCALE;
			mLeftBottomX = GameView.SCREEN_HALFWIDTH + (mLeftBottom.mX / scale);
			mLeftBottomY = Render.LOOK_Y - (mLeftBottom.mZ / scale);
	
			scale = (mRightTop.mY + (Render.CAMERA_FOV << Camera.PRECISION_LOG2)) / Render.SCALE;
			mRightTopX = GameView.SCREEN_HALFWIDTH + (mRightTop.mX / scale);
			mRightTopY = Render.LOOK_Y - (mRightTop.mZ / scale);
	
			scale = (mRightBottom.mY + (Render.CAMERA_FOV << Camera.PRECISION_LOG2)) / Render.SCALE;
			mRightBottomX = GameView.SCREEN_HALFWIDTH + (mRightBottom.mX / scale);
			mRightBottomY = Render.LOOK_Y - (mRightBottom.mZ / scale);
			//
			if( Render.sIsPreview ) {
				mLeftBottomX = mLeftTopX;
				mRightBottomX = mRightTopX;
			}
		}
	}
	//@Override
	public int compareTo(Object s) {
		return mViewDistance_d64 - ((Section)s).mViewDistance_d64;
	}
}