/** Track's data. Maximal track diameter must be less then 32768.
 * This limitation due horizon distance equals 65536
 */
public abstract class Track {
	public static final int TRACKS_CENTER_X = 16384;
	public static final int TRACKS_CENTER_Y = 16384;
	/** track's name */
	public char[] mName;
	/** color for start line on top side */
	public int mRoadStartLineColor;
	/** color for dark side lines on top */
	public int mRoadLineColorDark;
	/** color for light side lines on top */
	public int mRoadLineColorLight;
	/** Road width to interpolate a height */
	public final int mRoadWidth = 384;
	/** Scenery for the track */
	public Model[] mScenery = SceneryData.mMixed;
	/** Segment count, not implemented */
	//public int mNumTrackSegments;
	/** Number of track coordinates */
	public int mNumTrackCoords;
	/** Amount of nitrous oxide for standard league, sec */
	public int mStandardBoost;
	/** Amount of nitrous oxide for super league, sec */
	//public int mSuperBoost;
	public int mBestLapTimeMs10;
	public final char[] mBestLapPilot = new char[12];
	public int mBestRaceTimeMs10;
	public final char[] mBestRacePilot = new char[12];
	/** 3d coordinates for left and right top corners */
	public TrackPiece[] mPieces;
	public int mSpeedDeviation_fp8;// int8
	public int mMaxSpeed_fp8;

	protected ITrackAnimation mAnimation;

	public Track() {
		mRoadStartLineColor = 0xffffff;
		mRoadLineColorDark = 0x400000;
		mRoadLineColorLight = 0xe0e000;
		resetBestTime();
		clear();
	}
	public final void resetBestTime() {
		mBestLapTimeMs10 = 0x00ffffff;// < 7fffffff / 10
		mBestRaceTimeMs10 = 0x00ffffff;// < 7fffffff / 10
		for( int i = 0; i < mBestLapPilot.length; i++ ) mBestLapPilot[i] = '-';
		for( int i = 0; i < mBestRacePilot.length; i++ ) mBestRacePilot[i] = '-';
	}
	public static void setName(final char[] target, final char[] name) {
		int i = 0;
		for( ; i < name.length; i++ ) {
			target[i] = name[i];
		}
		for( ; i < target.length; i++ ) {
			target[i] = ' ';
		}
	}
	public final void clear() {
		mPieces = null;
	}
	/** calculate count of coordinates */
	protected final void updateData() {
		int coords_count = 0;
		for( int i = 0; i < mPieces.length; i++ ) {
			final TrackPiece tp = mPieces[i];
			coords_count += tp.mCoords.length;
			coords_count -= 2;
			// km/h to px/frame
			tp.mRecommendedSpeed_fp8 = tp.mRecommendedSpeed_fp8 * (256 * (int)GameView.REFRESH_PERIOD * 2 / 36);
		}
		mNumTrackCoords = coords_count;
		// km/h to px/frame
		mMaxSpeed_fp8 = mMaxSpeed_fp8 * (256 * (int)GameView.REFRESH_PERIOD * 2 / 36);
		mSpeedDeviation_fp8 = mSpeedDeviation_fp8 * (256 * (int)GameView.REFRESH_PERIOD * 2 / 36);
	}
	public abstract void init();
	/** scale down track's coordinates */
	/*public final Track scaleDown(int xscale, int yscale, int zscale) {
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				coords[i].mX /= xscale;
				coords[i].mY /= yscale;
				coords[i].mZ /= zscale;
			}
		}
		return this;
	}*/
	/** scale up track's coordinates */
	/*public final Track scaleUp(int xscale, int yscale, int zscale) {
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				coords[i].mX *= xscale;
				coords[i].mY *= yscale;
				coords[i].mZ *= zscale;
			}
		}
		return this;
	}*/
	/** shift track's coordinates */
	/*public final Track shift(int dx, int dy, int dz) {
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				coords[i].mX += dx;
				coords[i].mY += dy;
				coords[i].mZ += dz;
			}
		}
		return this;
	}*/
	/** to center XY track's coordinates around 0.0 */
	/*public final Track centerXYToZero() {
		int xmin = 32767;
		int xmax = 0;
		int ymin = 32767;
		int ymax = 0;
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				if( xmin > coords[i].mX ) xmin = coords[i].mX;
				if( xmax < coords[i].mX ) xmax = coords[i].mX;
				if( ymin > coords[i].mY ) ymin = coords[i].mY;
				if( ymax < coords[i].mY ) ymax = coords[i].mY;
			}
		}
		int midx = (xmax - xmin) >> 1;
		int midy = (ymax - ymin) >> 1;
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				coords[i].mX -= midx;
				coords[i].mY -= midy;
			}
		}
		return this;
	}*/
	public static final boolean isRoadSegmentBefore(final TrackPiece piece, int offset, int x, int y) {
		final Coord3D[] c = piece.mCoords;
		final Coord3D sc = c[offset];
		final Coord3D ec = c[offset + 1];
		// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
		// orientation > 0 : left
		// orientation < 0 : right
		// orientation == 0 : on
		return ( (ec.mX - sc.mX) * (y - sc.mY) + (sc.mY - ec.mY) * (x - sc.mX) < 0 );
	}
	public static final boolean isRoadSegmentBehind(final TrackPiece piece, int offset, int x, int y) {
		// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
		// orientation > 0 : left
		// orientation < 0 : right
		// orientation == 0 : on
		final Coord3D[] c = piece.mCoords;
		final Coord3D sc = c[offset];
		final Coord3D ec = c[offset + 1];
		return ( (ec.mX - sc.mX) * (y - sc.mY) + (sc.mY - ec.mY) * (x - sc.mX) > 0 );
	}
	public static final boolean isOnRoad(final TrackPiece piece, int offset, int x, int y) {
		// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
		// orientation > 0 : left
		// orientation < 0 : right
		// orientation == 0 : on
		int next_offset = offset + 2;
		final Coord3D[] c = piece.mCoords;
		final Coord3D left_near = c[offset];
		final Coord3D right_near = c[offset + 1];
		final Coord3D left_far = c[next_offset];
		final Coord3D right_far = c[next_offset + 1];
		//
		if( (left_far.mX - left_near.mX) * (y - left_near.mY) -
			(left_far.mY - left_near.mY) * (x - left_near.mX) <= 0 ) {
			if( (right_far.mX - right_near.mX) * (y - right_near.mY) -
				(right_far.mY - right_near.mY) * (x - right_near.mX) >= 0 )
				return true;
		}
		return false;
	}
	/** Convert cos^2 to angle from 0 to 90 degrees */
	private static final int COS_PRECISION_LOG2 = 8;
	/** Cosines to angle from 0 to 90 degrees */
	private static final int[] sCosToAngle = new int[] {// [(1 << COS_PRECISION_LOG2) + 1]
		16384, 15731, 15460, 15252, 15076, 14921, 14780, 14651, 14530, 14416,
		14308, 14206, 14107, 14013, 13922, 13833, 13748, 13665, 13584, 13506,
		13429, 13354, 13280, 13208, 13138, 13068, 13000, 12934, 12868, 12803,
		12739, 12676, 12614, 12553, 12493, 12433, 12374, 12316, 12258, 12201,
		12145, 12089, 12034, 11979, 11925, 11871, 11818, 11765, 11712, 11660,
		11609, 11558, 11507, 11456, 11406, 11357, 11307, 11258, 11209, 11161,
		11112, 11064, 11017, 10969, 10922, 10875, 10829, 10782, 10736, 10690,
		10644, 10598, 10553, 10508, 10463, 10418, 10373, 10329, 10284, 10240,
		10196, 10152, 10109, 10065, 10022,  9978,  9935,  9892,  9849,  9806,
		 9763,  9721,  9678,  9636,  9594,  9551,  9509,  9467,  9425,  9383,
		 9342,  9300,  9258,  9217,  9175,  9134,  9092,  9051,  9010,  8969,
		 8927,  8886,  8845,  8804,  8763,  8722,  8681,  8640,  8599,  8558,
		 8518,  8477,  8436,  8395,  8355,  8314,  8273,  8232,  8192,  8151,
		 8110,  8069,  8028,  7988,  7947,  7906,  7865,  7825,  7784,  7743,
		 7702,  7661,  7620,  7579,  7538,  7497,  7456,  7414,  7373,  7332,
		 7291,  7249,  7208,  7166,  7125,  7083,  7041,  7000,  6958,  6916,
		 6874,  6832,  6789,  6747,  6705,  6662,  6620,  6577,  6534,  6491,
		 6448,  6405,  6361,  6318,  6274,  6231,  6187,  6143,  6099,  6054,
		 6010,  5965,  5920,  5875,  5830,  5785,  5739,  5693,  5647,  5601,
		 5554,  5508,  5461,  5414,  5366,  5319,  5271,  5222,  5174,  5125,
		 5076,  5026,  4977,  4927,  4876,  4825,  4774,  4723,  4671,  4618,
		 4565,  4512,  4458,  4404,  4349,  4294,  4238,  4182,  4125,  4067,
		 4009,  3950,  3890,  3830,  3769,  3707,  3644,  3580,  3515,  3449,
		 3383,  3315,  3245,  3175,  3103,  3029,  2954,  2877,  2799,  2718,
		 2635,  2550,  2461,  2370,  2276,  2177,  2075,  1967,  1853,  1732,
		 1603, 1462, 1307, 1131, 923, 652, 0
	};
	/**
	 * 
	 * @param piece
	 * @param offset
	 * @return angle from -180 degrees to +180 degrees
	 */
	public static int getSectionAngle(final TrackPiece piece, int offset) {
		final Coord3D[] c = piece.mCoords;
		final Coord3D near_left = c[offset++];// c[offset];
		final Coord3D near_right = c[offset++];// c[offset + 1];
		final Coord3D far_left = c[offset++];// c[next_offset];
		final Coord3D far_right = c[offset];// c[next_offset + 1];
		//
		final int dx = ((far_left.mX + far_right.mX) - (near_left.mX + near_right.mX)) >> 1;
		final int dy = ((far_left.mY + far_right.mY) - (near_left.mY + near_right.mY)) >> 1;
		// cos^2 = (x1 * x2 + y1 * y2)/((x1 * x1 + y1 * y1) * (x2 * x2 + y2 * y2))
		// for base vector x1 = 0, y1 = 1, so:
		// cos^2 = (y2 * y2)/(x2 * x2 + y2 * y2)
		final int dy2 = dy * dy;
		final int cos2 = (dy2 << COS_PRECISION_LOG2) / (dx * dx + dy2);
		int angle = sCosToAngle[Math.abs( cos2 )];
		if( dy < 0 ) angle = Camera.ANGLE_180 - angle;
		if( dx < 0 ) return -angle;
		return angle;
	}
}